import logging
import math
import multiprocessing as mp
import numpy as np
import time

from typing import List, Callable
from flare.kernels.utils import from_mask_to_args, from_grad_to_mask

_global_training_data = {}
_global_training_labels = {}
_global_training_structures = {}
_global_energy_labels = {}


def queue_wrapper(result_queue, wid,
                  func, args):
    """
    wrapper function for multiprocessing queue
    """
    result_queue.put((wid, func(*args)))


def partition_matrix(n_sample, size, n_cpus):
    """
    partition the training data for matrix calculation
    the number of blocks are close to n_cpus
    since mp.Process does not allow to change the thread number
    """

    # divide the block by n_cpu partitions, with size n_sample0
    # if the divided chunk is smaller than the requested chunk n_sample
    # use the requested chunk size
    n_unique_elements = size * (size + 1) / 2
    n_sample0 = int(math.ceil(np.sqrt(n_unique_elements / n_cpus)))
    if (n_sample0 > n_sample):
        n_sample = n_sample0

    block_id = []
    nbatch = 0
    nbatch1 = 0
    nbatch2 = 0
    e1 = 0
    while (e1 < size):
        s1 = int(n_sample*nbatch1)
        e1 = int(np.min([s1 + n_sample, size]))
        nbatch2 = nbatch1
        nbatch1 += 1
        e2 = 0
        while (e2 < size):
            s2 = int(n_sample * nbatch2)
            e2 = int(np.min([s2 + n_sample, size]))
            block_id += [(s1, e1, s2, e2)]
            nbatch2 += 1
            nbatch += 1

    return block_id, nbatch


def partition_matrix_custom(n_sample: int, start1, end1, start2, end2, n_cpus):
    """
    Partition a specified portion of a matrix.
    """

    n_rows = end1 - start1
    n_cols = end2 - start2
    n_unique_elements = n_rows * n_cols
    n_sample0 = int(math.ceil(np.sqrt(n_unique_elements / n_cpus)))
    if (n_sample0 > n_sample):
        n_sample = n_sample0

    block_id = []
    n_block = 0  # total number of blocks
    nbatch1 = 0
    nbatch2 = 0
    e1 = 0
    while (e1 < end1):
        s1 = start1 + n_sample * nbatch1
        e1 = np.min([s1 + n_sample, end1])
        nbatch1 += 1

        e2 = 0
        nbatch2 = 0
        while (e2 < end2):
            s2 = start2 + n_sample * nbatch2
            e2 = np.min([s2 + n_sample, end2])
            block_id += [(s1, e1, s2, e2)]
            nbatch2 += 1
            n_block += 1  # update block count

    return block_id, n_block


def partition_vector(n_sample, size, n_cpus):
    """
    partition the training data for vector calculation
    the number of blocks are the same as n_cpus
    since mp.Process does not allow to change the thread number
    """
    n_sample0 = int(math.ceil(size/n_cpus))
    if (n_sample0 > n_sample):
        n_sample = n_sample0

    block_id = []
    nbatch = 0
    e = 0
    while (e < size):
        s = n_sample*nbatch
        e = np.min([s + n_sample, size])
        block_id += [(s, e)]
        nbatch += 1
    return block_id, nbatch


def partition_force_energy_block(n_sample: int, size1: int, size2: int,
                                 n_cpus: int):
    """Special partition method for the force/energy block. Because the number
    of environments in a structure can vary, we only split up the environment
    list, which has length size1.

    Note that two sizes need to be specified: the size of the envionment
    list and the size of the structure list.

    Args:
        n_sample (int): Number of environments per processor.
        size1 (int): Size of the environment list.
        size2 (int): Size of the structure list.
        n_cpus (int): Number of cpus.
    """

    n_sample0 = int(math.ceil(size1/n_cpus))
    if (n_sample0 > n_sample):
        n_sample = n_sample0

    block_id = []
    nbatch = 0
    e = 0

    while (e < size1):
        s = n_sample * nbatch
        e = np.min([s + n_sample, size1])
        block_id += [(s, e, 0, size2)]
        nbatch += 1

    return block_id, nbatch


def partition_update(n_sample, size, old_size, n_cpus):

    n_unique_elements = (size - old_size) * size
    n_sample0 = int(math.ceil(np.sqrt(n_unique_elements / n_cpus)))
    if (n_sample0 > n_sample):
        n_sample = n_sample0

    ns_new = int(math.ceil((size - old_size) / n_sample))
    old_ns = int(math.ceil(old_size / n_sample))

    nbatch = 0
    block_id = []
    for ibatch1 in range(old_ns):
        s1 = int(n_sample*ibatch1)
        e1 = int(np.min([s1 + n_sample, old_size]))

        for ibatch2 in range(ns_new):
            s2 = int(n_sample * ibatch2) + old_size
            e2 = int(np.min([s2 + n_sample, size]))
            block_id += [(s1, e1, s2, e2)]
            nbatch += 1

    for ibatch1 in range(ns_new):
        s1 = int(n_sample * ibatch1) + old_size
        e1 = int(np.min([s1 + n_sample, size]))

        for ibatch2 in range(ns_new):
            s2 = int(n_sample * ibatch2) + old_size
            e2 = int(np.min([s2 + n_sample, size]))
            block_id += [(s1, e1, s2, e2)]
            nbatch += 1

    return block_id, nbatch


def obtain_noise_len(hyps, hyps_mask):
    """
    obtain the noise parameter from hyps and mask
    """

    # assume sigma_n is the final hyperparameter
    sigma_n = hyps[-1]
    # correct it if map is defined
    non_noise_hyps = len(hyps)-1
    train_noise = True
    if (hyps_mask is not None):
        train_noise = hyps_mask.get('train_noise', True)
        if (train_noise is False):
            sigma_n = hyps_mask['original_hyps'][-1]
            non_noise_hyps = len(hyps)

    return sigma_n, non_noise_hyps, train_noise


# --------------------------------------------------------------------------
#                   Parallel matrix/vector construction
# --------------------------------------------------------------------------

def parallel_matrix_construction(pack_function, hyps, name, kernel, cutoffs,
                                 hyps_mask, block_id, nbatch, size1, size2,
                                 m1, m2, symm=True):
    result_queue = mp.Queue()
    children = []
    for wid in range(nbatch):
        s1, e1, s2, e2 = block_id[wid]
        children.append(mp.Process(target=queue_wrapper,
                                   args=(result_queue, wid, pack_function,
                                         (hyps, name, s1, e1, s2, e2, s1 == s2,
                                          kernel, cutoffs, hyps_mask))))

    # Run child processes.
    for c in children:
        c.start()

    # Wait for all results to arrive.
    matrix = np.zeros((size1, size2))
    for _ in range(nbatch):
        wid, result_chunk = result_queue.get(block=True)
        s1, e1, s2, e2 = block_id[wid]
        matrix[s1 * m1:e1 * m1,
               s2 * m2:e2 * m2] = result_chunk
        # Note that the force/energy block is not symmetric; in this case
        # symm is False.
        if ((s1 != s2) and (symm is True)):
            matrix[s2 * m2:e2 * m2, s1 * m1:e1 * m1] = result_chunk.T

    # Join child processes (clean up zombies).
    for c in children:
        c.join()

    # matrix manipulation
    del result_queue
    del children

    return matrix


def parallel_vector_construction(pack_function, name, x, kernel, hyps,
                                 cutoffs, hyps_mask, block_id, nbatch, size,
                                 mult, d_1=None):
    result_queue = mp.Queue()
    children = []
    for wid in range(nbatch):
        s, e = block_id[wid]
        children.append(
            mp.Process(
                target=queue_wrapper,
                args=(result_queue, wid, pack_function,
                      (name, s, e, x, kernel, hyps, cutoffs, hyps_mask, d_1))))

    # Run child processes.
    for c in children:
        c.start()

    # Wait for all results to arrive.
    vector = np.zeros(size * mult)
    for _ in range(nbatch):
        wid, result_chunk = result_queue.get(block=True)
        s, e = block_id[wid]
        vector[s * mult:e * mult] = result_chunk

    # Join child processes (clean up zombies).
    for c in children:
        c.join()

    return vector


def multiple_array_construction(pack_function, name, x, kernel, hyps,
                                cutoffs, hyps_mask, block_id, nbatch, size,
                                mult, array_sizes):
    result_queue = mp.Queue()
    children = []
    for wid in range(nbatch):
        s, e = block_id[wid]
        children.append(
            mp.Process(
                target=queue_wrapper,
                args=(result_queue, wid, pack_function,
                      (name, s, e, x, kernel, hyps, cutoffs, hyps_mask))))

    # Run child processes.
    for c in children:
        c.start()

    # Initialize arrays.
    n_arrays = len(array_sizes)
    arrays = []
    for n in range(n_arrays):
        arrays.append(np.zeros(array_sizes[n]))

    # Wait for all results to arrive.
    for _ in range(nbatch):
        wid, result_chunk = result_queue.get(block=True)
        s, e = block_id[wid]
        for n in range(n_arrays):
            arrays[n][:, s * mult:e * mult] = result_chunk[n]

    # Join child processes (clean up zombies).
    for c in children:
        c.join()

    return arrays


# --------------------------------------------------------------------------
#                            Ky construction
# --------------------------------------------------------------------------

def get_force_block_pack(hyps: np.ndarray, name: str, s1: int, e1: int,
                         s2: int, e2: int, same: bool, kernel, cutoffs,
                         hyps_mask):
    """ Compute covariance matrix element between set1 and set2
    :param hyps: list of hyper-parameters
    :param name: name of the gp instance.
    :param same: whether the row and column are the same
    :param kernel: function object of the kernel
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    :return: covariance matrix
    """

    # initialize matrices
    training_data = _global_training_data[name]
    size1 = (e1-s1)*3
    size2 = (e2-s2)*3
    force_block = np.zeros([size1, size2])

    ds = [1, 2, 3]

    # calculate elements
    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    for m_index in range(size1):
        x_1 = training_data[int(math.floor(m_index / 3))+s1]
        d_1 = ds[m_index % 3]
        if (same):
            lowbound = m_index
        else:
            lowbound = 0
        for n_index in range(lowbound, size2):
            x_2 = training_data[int(math.floor(n_index / 3))+s2]
            d_2 = ds[n_index % 3]
            kern_curr = kernel(x_1, x_2, d_1, d_2, *args)
            # store kernel value
            force_block[m_index, n_index] = kern_curr
            if (same):
                force_block[n_index, m_index] = kern_curr

    return force_block


def get_energy_block_pack(hyps: np.ndarray, name: str, s1: int, e1: int,
                          s2: int, e2: int, same: bool, kernel, cutoffs,
                          hyps_mask):

    # initialize matrices
    training_structures = _global_training_structures[name]
    size1 = e1 - s1
    size2 = e2 - s2
    energy_block = np.zeros([size1, size2])

    # calculate elements
    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    for m_index in range(size1):
        struc_1 = training_structures[m_index + s1]
        if (same):
            lowbound = m_index
        else:
            lowbound = 0

        for n_index in range(lowbound, size2):
            struc_2 = training_structures[n_index + s2]

            # Loop over environments in both structures to compute the
            # energy/energy kernel.
            kern_curr = 0
            for environment_1 in struc_1:
                for environment_2 in struc_2:
                    kern_curr += kernel(environment_1, environment_2, *args)

            # Store kernel value.
            energy_block[m_index, n_index] = kern_curr
            if (same):
                energy_block[n_index, m_index] = kern_curr

    return energy_block


def get_force_energy_block_pack(hyps: np.ndarray, name: str, s1: int,
                                e1: int, s2: int, e2: int, same: bool, kernel,
                                cutoffs, hyps_mask):
    # initialize matrices
    training_data = _global_training_data[name]
    training_structures = _global_training_structures[name]
    size1 = (e1 - s1) * 3
    size2 = e2 - s2
    force_energy_block = np.zeros([size1, size2])

    ds = [1, 2, 3]

    # calculate elements
    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    for m_index in range(size1):
        environment_1 = training_data[int(math.floor(m_index / 3)) + s1]
        d_1 = ds[m_index % 3]

        for n_index in range(size2):
            structure = training_structures[n_index + s2]

            # Loop over environments in the training structure.
            kern_curr = 0
            for environment_2 in structure:
                kern_curr += kernel(environment_1, environment_2, d_1, *args)

            # store kernel value
            force_energy_block[m_index, n_index] = kern_curr

    return force_energy_block


def get_force_block(hyps: np.ndarray, name: str, kernel, cutoffs=None,
                    hyps_mask=None, n_cpus=1, n_sample=100):
    """ parallel version of get_ky_mat
    :param hyps: list of hyper-parameters
    :param name: name of the gp instance.
    :param kernel: function object of the kernel
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    :return: covariance matrix
    """

    training_data = _global_training_data[name]
    size = len(training_data)
    size3 = 3 * size

    if (n_cpus is None):
        n_cpus = mp.cpu_count()
    if (n_cpus == 1):
        k_mat = \
            get_force_block_pack(hyps, name, 0, size, 0, size, True, kernel,
                                 cutoffs, hyps_mask)
    else:
        # initialize matrices
        block_id, nbatch = partition_matrix(n_sample, size, n_cpus)
        mult = 3

        k_mat = \
            parallel_matrix_construction(get_force_block_pack, hyps, name,
                                         kernel, cutoffs, hyps_mask,
                                         block_id, nbatch, size3, size3,
                                         mult, mult)

    sigma_n, _, __ = obtain_noise_len(hyps, hyps_mask)
    force_block = k_mat
    force_block += sigma_n ** 2 * np.eye(size3)

    return force_block


def get_energy_block(hyps: np.ndarray, name: str, kernel, energy_noise,
                     cutoffs=None, hyps_mask=None, n_cpus=1, n_sample=100):
    training_structures = _global_training_structures[name]
    size = len(training_structures)

    if (n_cpus is None):
        n_cpus = mp.cpu_count()
    if (n_cpus == 1):
        k_mat = \
            get_energy_block_pack(hyps, name, 0, size, 0, size, True, kernel,
                                  cutoffs, hyps_mask)
    else:
        block_id, nbatch = partition_matrix(n_sample, size, n_cpus)
        mult = 1

        k_mat = \
            parallel_matrix_construction(get_energy_block_pack, hyps, name,
                                         kernel, cutoffs, hyps_mask,
                                         block_id, nbatch, size, size, mult,
                                         mult)

    energy_block = k_mat
    energy_block += (energy_noise ** 2) * np.eye(size)

    return energy_block


def get_force_energy_block(hyps: np.ndarray, name: str, kernel, cutoffs=None,
                           hyps_mask=None, n_cpus=1, n_sample=100):
    training_data = _global_training_data[name]
    training_structures = _global_training_structures[name]
    size1 = len(training_data) * 3
    size2 = len(training_structures)
    size3 = len(training_data)

    if (n_cpus is None):
        n_cpus = mp.cpu_count()
    if (n_cpus == 1):
        force_energy_block = \
            get_force_energy_block_pack(hyps, name, 0, size3, 0, size2, True,
                                        kernel, cutoffs, hyps_mask)
    else:
        # initialize matrices
        block_id, nbatch = \
            partition_force_energy_block(n_sample, size3, size2, n_cpus)
        m1 = 3  # 3 force components per environment
        m2 = 1  # 1 energy per structure

        force_energy_block = \
            parallel_matrix_construction(get_force_energy_block_pack, hyps,
                                         name, kernel, cutoffs, hyps_mask,
                                         block_id, nbatch, size1, size2, m1,
                                         m2, symm=False)

    return force_energy_block


def get_Ky_mat(hyps: np.ndarray, name: str, force_kernel: Callable,
               energy_kernel: Callable, force_energy_kernel: Callable,
               energy_noise, cutoffs=None, hyps_mask=None,
               n_cpus=1, n_sample=100):

    training_data = _global_training_data[name]
    training_structures = _global_training_structures[name]
    size1 = len(training_data) * 3
    size2 = len(training_structures)

    # Initialize Ky.
    ky_mat = np.zeros((size1 + size2, size1 + size2))

    # Assemble the full covariance matrix block-by-block.
    force_block = get_force_block(hyps, name, force_kernel, cutoffs, hyps_mask,
                                  n_cpus, n_sample)

    energy_block = get_energy_block(hyps, name, energy_kernel, energy_noise,
                                    cutoffs, hyps_mask, n_cpus, n_sample)

    force_energy_block = \
        get_force_energy_block(hyps, name, force_energy_kernel, cutoffs,
                               hyps_mask, n_cpus, n_sample)

    ky_mat[0:size1, 0:size1] = force_block
    ky_mat[size1:, size1:] = energy_block
    ky_mat[0:size1, size1:] = force_energy_block
    ky_mat[size1:, 0:size1] = force_energy_block.transpose()

    return ky_mat


# --------------------------------------------------------------------------
#                              Ky updates
# --------------------------------------------------------------------------

def update_force_block(ky_mat_old: np.ndarray, n_envs_prev: int,
                       hyps: np.ndarray, name: str, kernel, cutoffs=None,
                       hyps_mask=None, n_cpus=1, n_sample=100):

    old_size = n_envs_prev
    old_size3 = n_envs_prev * 3
    mult = 3
    size = len(_global_training_data[name])
    size3 = 3 * size

    if (n_cpus is None):
        n_cpus = mp.cpu_count()

    # serial version
    if (n_cpus == 1):
        force_block = np.zeros((size3, size3))
        new_mat = \
            get_force_block_pack(hyps, name, 0, size, old_size, size, False,
                                 kernel, cutoffs, hyps_mask)
        force_block[:, old_size3:] = new_mat
        force_block[old_size3:, :] = new_mat.transpose()

    # parallel version
    else:
        block_id, nbatch = partition_update(n_sample, size, old_size, n_cpus)

        force_block = \
            parallel_matrix_construction(get_force_block_pack, hyps, name,
                                         kernel, cutoffs, hyps_mask,
                                         block_id, nbatch, size3, size3,
                                         mult, mult)

    # insert previous covariance matrix
    force_block[:old_size3, :old_size3] = \
        ky_mat_old[:old_size3, :old_size3]

    # add the noise parameter
    sigma_n, _, _ = obtain_noise_len(hyps, hyps_mask)
    force_block[old_size3:, old_size3:] += \
        sigma_n ** 2 * np.eye(size3-old_size3)

    return force_block


def update_energy_block(ky_mat_old: np.ndarray, n_envs_prev: int,
                        hyps: np.ndarray, name: str, kernel,
                        energy_noise: float, cutoffs=None, hyps_mask=None,
                        n_cpus=1, n_sample=100):

    old_size = ky_mat_old.shape[0] - 3 * n_envs_prev
    mult = 1
    size = len(_global_training_structures[name])

    if (n_cpus is None):
        n_cpus = mp.cpu_count()

    # serial version
    if (n_cpus == 1):
        energy_block = np.zeros((size, size))
        new_mat = \
            get_energy_block_pack(hyps, name, 0, size, old_size, size, False,
                                  kernel, cutoffs, hyps_mask)
        energy_block[:, old_size:] = new_mat
        energy_block[old_size:, :] = new_mat.transpose()

    # parallel version
    else:
        block_id, nbatch = partition_update(n_sample, size, old_size, n_cpus)

        energy_block = \
            parallel_matrix_construction(get_energy_block_pack, hyps, name,
                                         kernel, cutoffs, hyps_mask,
                                         block_id, nbatch, size, size,
                                         mult, mult)

    # insert previous covariance matrix (if it has nonzero size)
    if old_size > 0:
        energy_block[:old_size, :old_size] = ky_mat_old[-old_size:, -old_size:]

    # add the noise parameter
    energy_block[old_size:, old_size:] += \
        (energy_noise ** 2) * np.eye(size - old_size)

    return energy_block


def update_force_energy_block(ky_mat_old: np.ndarray, n_envs_prev: int,
                              hyps: np.ndarray, name: str, kernel,
                              energy_noise: float, cutoffs=None,
                              hyps_mask=None, n_cpus=1, n_sample=100):

    n_strucs_prev = ky_mat_old.shape[0] - 3 * n_envs_prev
    n_envs = len(_global_training_data[name])
    n_strucs = len(_global_training_structures[name])
    force_energy_block = np.zeros((n_envs * 3, n_strucs))

    if (n_cpus is None):
        n_cpus = mp.cpu_count()

    # serial version
    if (n_cpus == 1):
        mat_1 = \
            get_force_energy_block_pack(hyps, name, 0, n_envs,
                                        n_strucs_prev, n_strucs,
                                        False, kernel, cutoffs, hyps_mask)

        mat_2 = \
            get_force_energy_block_pack(hyps, name, n_envs_prev, n_envs,
                                        0, n_strucs_prev,
                                        False, kernel, cutoffs, hyps_mask)

    # parallel version
    else:
        force_energy_block = np.zeros((n_envs * 3, n_strucs))

        block_id_1, nbatch_1 = \
            partition_matrix_custom(n_sample, 0, n_envs, n_strucs_prev,
                                    n_strucs, n_cpus)
        block_id_2, nbatch_2 = \
            partition_matrix_custom(n_sample, n_envs_prev, n_envs,
                                    0, n_strucs_prev, n_cpus)

        size1 = n_envs * 3
        size2 = n_strucs
        m1 = 3
        m2 = 1

        mat_1 = \
            parallel_matrix_construction(get_force_energy_block_pack,
                                         hyps, name, kernel, cutoffs,
                                         hyps_mask, block_id_1, nbatch_1,
                                         size1, size2, m1, m2, False)
        mat_2 = \
            parallel_matrix_construction(get_force_energy_block_pack,
                                         hyps, name, kernel, cutoffs,
                                         hyps_mask, block_id_2, nbatch_2,
                                         size1, size2, m1, m2, False)

        # reduce the size of the matrices
        mat_1 = mat_1[:n_envs * 3, n_strucs_prev:n_strucs]
        mat_2 = mat_2[n_envs_prev * 3:n_envs * 3, :n_strucs_prev]

    force_energy_block[:n_envs * 3, n_strucs_prev:n_strucs] = mat_1
    force_energy_block[n_envs_prev * 3:n_envs * 3, :n_strucs_prev] = mat_2

    # insert previous covariance matrix
    force_energy_block[:n_envs_prev * 3, :n_strucs_prev] = \
        ky_mat_old[:n_envs_prev * 3,
                   n_envs_prev * 3:n_envs_prev * 3 + n_strucs_prev]

    return force_energy_block


def get_ky_mat_update(ky_mat_old: np.ndarray, n_envs_prev: int,
                      hyps: np.ndarray, name: str, force_kernel: Callable,
                      energy_kernel: Callable, force_energy_kernel: Callable,
                      energy_noise: float, cutoffs=None, hyps_mask=None,
                      n_cpus=1, n_sample=100):

    n_envs = len(_global_training_data[name])
    n_strucs = len(_global_training_structures[name])
    size1 = n_envs * 3
    size2 = n_strucs
    ky_mat = np.zeros((size1 + size2, size1 + size2))

    force_block = \
        update_force_block(ky_mat_old, n_envs_prev, hyps, name, force_kernel,
                           cutoffs, hyps_mask, n_cpus, n_sample)

    energy_block = \
        update_energy_block(ky_mat_old, n_envs_prev, hyps, name, energy_kernel,
                            energy_noise, cutoffs, hyps_mask, n_cpus, n_sample)

    force_energy_block = \
        update_force_energy_block(ky_mat_old, n_envs_prev, hyps, name,
                                  force_energy_kernel, energy_noise, cutoffs,
                                  hyps_mask, n_cpus, n_sample)

    ky_mat[0:size1, 0:size1] = force_block
    ky_mat[size1:, size1:] = energy_block
    ky_mat[0:size1, size1:] = force_energy_block
    ky_mat[size1:, 0:size1] = force_energy_block.transpose()

    return ky_mat


# --------------------------------------------------------------------------
#                            Kernel vectors
# --------------------------------------------------------------------------

def energy_energy_vector_unit(name, s, e, x, kernel, hyps, cutoffs=None,
                              hyps_mask=None, d_1=None):
    """
    Gets part of the energy/energy vector.
    """

    training_structures = _global_training_structures[name]

    size = e - s
    energy_energy_unit = np.zeros(size, )

    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    for m_index in range(size):
        structure = training_structures[m_index + s]
        kern_curr = 0
        for environment in structure:
            kern_curr += kernel(x, environment, *args)

        energy_energy_unit[m_index] = kern_curr

    return energy_energy_unit


def energy_force_vector_unit(name, s, e, x, kernel, hyps, cutoffs=None,
                             hyps_mask=None, d_1=None):
    """
    Gets part of the energy/force vector.
    """

    training_data = _global_training_data[name]

    ds = [1, 2, 3]
    size = (e - s) * 3
    k_v = np.zeros(size, )

    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    for m_index in range(size):
        x_2 = training_data[int(math.floor(m_index / 3))+s]
        d_2 = ds[m_index % 3]
        k_v[m_index] = kernel(x_2, x, d_2, *args)

    return k_v


def force_energy_vector_unit(name, s, e, x, kernel, hyps, cutoffs, hyps_mask,
                             d_1):
    """
    Gets part of the force/energy vector.
    """

    size = e - s
    args = from_mask_to_args(hyps, cutoffs, hyps_mask)
    force_energy_unit = np.zeros(size,)

    for m_index in range(size):
        training_structure = _global_training_structures[name][m_index+s]
        kern_curr = 0
        for environment in training_structure:
            kern_curr += kernel(x, environment, d_1, *args)

        force_energy_unit[m_index] = kern_curr

    return force_energy_unit


def force_force_vector_unit(name, s, e, x, kernel, hyps, cutoffs, hyps_mask,
                            d_1):
    """
    Gets part of the force/force vector.
    """

    size = e - s
    ds = [1, 2, 3]

    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    k_v = np.zeros(size * 3)

    for m_index in range(size):
        x_2 = _global_training_data[name][m_index + s]
        for d_2 in ds:
            k_v[m_index * 3 + d_2 - 1] = kernel(x, x_2, d_1, d_2, *args)

    return k_v


def efs_force_vector_unit(name, s, e, x, efs_force_kernel, hyps, cutoffs,
                          hyps_mask):

    training_data = _global_training_data[name]

    size = e - s
    k_ef = np.zeros((1, size * 3))
    k_ff = np.zeros((3, size * 3))
    k_sf = np.zeros((6, size * 3))

    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    for m_index in range(size):
        x_2 = training_data[m_index + s]
        ef, ff, sf = efs_force_kernel(x, x_2, *args)

        ind1 = m_index * 3
        ind2 = (m_index + 1) * 3

        k_ef[:, ind1:ind2] = ef
        k_ff[:, ind1:ind2] = ff
        k_sf[:, ind1:ind2] = sf

    return k_ef, k_ff, k_sf


def efs_energy_vector_unit(name, s, e, x, efs_energy_kernel, hyps, cutoffs,
                           hyps_mask):

    size = e - s
    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    k_ee = np.zeros((1, size))
    k_fe = np.zeros((3, size))
    k_se = np.zeros((6, size))

    for m_index in range(size):
        training_structure = _global_training_structures[name][m_index + s]

        ee_curr = 0
        fe_curr = np.zeros(3)
        se_curr = np.zeros(6)

        for environment in training_structure:
            ee, fe, se = efs_energy_kernel(x, environment, *args)
            ee_curr += ee
            fe_curr += fe
            se_curr += se

        k_ee[:, m_index] = ee_curr
        k_fe[:, m_index] = fe_curr
        k_se[:, m_index] = se_curr

    return k_ee, k_fe, k_se


def energy_energy_vector(name, kernel, x, hyps, cutoffs=None,
                         hyps_mask=None, n_cpus=1, n_sample=100):
    """
    Get a vector of covariances between the local energy of a test environment
    and the total energy labels in the training set.
    """

    size = len(_global_training_structures[name])

    if (n_cpus is None):
        n_cpus = mp.cpu_count()
    if (n_cpus == 1):
        return energy_energy_vector_unit(name, 0, size, x, kernel, hyps,
                                         cutoffs, hyps_mask)

    block_id, nbatch = partition_vector(n_sample, size, n_cpus)
    pack_function = energy_energy_vector_unit
    mult = 1

    force_energy_vector = \
        parallel_vector_construction(pack_function, name, x, kernel,
                                     hyps, cutoffs, hyps_mask, block_id,
                                     nbatch, size, mult)

    return force_energy_vector


def energy_force_vector(name, kernel, x, hyps, cutoffs=None, hyps_mask=None,
                        n_cpus=1, n_sample=100):
    """
    Get the vector of covariances between the local energy of a test
    environment and the force labels in the training set.
    """

    training_data = _global_training_data[name]
    size = len(training_data)

    if (n_cpus is None):
        n_cpus = mp.cpu_count()
    if (n_cpus == 1):
        return energy_force_vector_unit(name, 0, size, x, kernel, hyps,
                                        cutoffs, hyps_mask)

    block_id, nbatch = partition_vector(n_sample, size, n_cpus)
    pack_function = energy_force_vector_unit
    mult = 3

    k12_v = parallel_vector_construction(pack_function, name, x, kernel,
                                         hyps, cutoffs, hyps_mask, block_id,
                                         nbatch, size, mult)

    return k12_v


def force_energy_vector(name, kernel, x, d_1, hyps, cutoffs=None,
                        hyps_mask=None, n_cpus=1, n_sample=100):
    """
    Get a vector of covariances between a force component of a test environment
    and the total energy labels in the training set.
    """

    size = len(_global_training_structures[name])

    if (n_cpus is None):
        n_cpus = mp.cpu_count()
    if (n_cpus == 1):
        return force_energy_vector_unit(name, 0, size, x, kernel, hyps,
                                        cutoffs, hyps_mask, d_1)

    block_id, nbatch = partition_vector(n_sample, size, n_cpus)
    pack_function = force_energy_vector_unit
    mult = 1

    force_energy_vector = \
        parallel_vector_construction(pack_function, name, x, kernel,
                                     hyps, cutoffs, hyps_mask, block_id,
                                     nbatch, size, mult, d_1)

    return force_energy_vector


def force_force_vector(name, kernel, x, d_1, hyps, cutoffs=None,
                       hyps_mask=None, n_cpus=1, n_sample=100):
    """
    Get a vector of covariances between a force component of a test environment
    and the force labels in the training set.
    """

    size = len(_global_training_data[name])

    if (n_cpus is None):
        n_cpus = mp.cpu_count()
    if (n_cpus == 1):
        return force_force_vector_unit(name, 0, size, x, kernel, hyps, cutoffs,
                                       hyps_mask, d_1)

    block_id, nbatch = partition_vector(n_sample, size, n_cpus)
    pack_function = force_force_vector_unit
    mult = 3

    k12_v = parallel_vector_construction(pack_function, name, x, kernel,
                                         hyps, cutoffs, hyps_mask, block_id,
                                         nbatch, size, mult, d_1)

    return k12_v


def efs_force_vector(name, efs_force_kernel, x, hyps, cutoffs=None,
                     hyps_mask=None, n_cpus=1, n_sample=100):
    """
    Returns covariances between the local eneregy, force components, and
    partial stresses of a test environment and the force labels in the
    training set.
    """

    size = len(_global_training_data[name])

    if (n_cpus is None):
        n_cpus = mp.cpu_count()

    # Perform serial calculation if n_cpus = 1.
    if (n_cpus == 1):
        return efs_force_vector_unit(name, 0, size, x, efs_force_kernel, hyps,
                                     cutoffs, hyps_mask)

    # Otherwise, perform parallel calculation.
    block_id, nbatch = partition_vector(n_sample, size, n_cpus)
    pack_function = efs_force_vector_unit
    mult = 3
    n_comps = size * mult
    array_sizes = [(1, n_comps), (3, n_comps), (6, n_comps)]

    efs_arrays = \
        multiple_array_construction(pack_function, name, x, efs_force_kernel,
                                    hyps, cutoffs, hyps_mask, block_id, nbatch,
                                    size, mult, array_sizes)

    return efs_arrays


def efs_energy_vector(name, efs_energy_kernel, x, hyps, cutoffs=None,
                      hyps_mask=None, n_cpus=1, n_sample=100):
    """
    Returns covariances between the local eneregy, force components, and
    partial stresses of a test environment and the total energy labels in the
    training set.
    """

    size = len(_global_training_structures[name])

    if (n_cpus is None):
        n_cpus = mp.cpu_count()

    # Perform serial calculation if n_cpus = 1.
    if (n_cpus == 1):
        return efs_energy_vector_unit(name, 0, size, x, efs_energy_kernel,
                                      hyps, cutoffs, hyps_mask)

    # Otherwise, perform parallel calculation.
    block_id, nbatch = partition_vector(n_sample, size, n_cpus)
    pack_function = efs_energy_vector_unit
    mult = 1
    array_sizes = [(1, size), (3, size), (6, size)]

    efs_arrays = \
        multiple_array_construction(pack_function, name, x, efs_energy_kernel,
                                    hyps, cutoffs, hyps_mask, block_id, nbatch,
                                    size, mult, array_sizes)

    return efs_arrays


def get_kernel_vector(name, force_force_kernel: Callable,
                      force_energy_kernel: Callable, x, d_1, hyps,
                      cutoffs=None, hyps_mask=None, n_cpus=1, n_sample=100):

    size1 = len(_global_training_data[name])
    size2 = len(_global_training_structures[name])
    kernel_vector = np.zeros(size1 * 3 + size2)

    force_vector = \
        force_force_vector(name, force_force_kernel, x, d_1, hyps, cutoffs,
                           hyps_mask, n_cpus, n_sample)
    energy_vector = \
        force_energy_vector(name, force_energy_kernel, x, d_1, hyps, cutoffs,
                            hyps_mask, n_cpus, n_sample)

    kernel_vector[0:size1*3] = force_vector
    kernel_vector[size1*3:] = energy_vector

    return kernel_vector


def en_kern_vec(name, energy_force_kernel, energy_energy_kernel, x, hyps,
                cutoffs=None, hyps_mask=None, n_cpus=1, n_sample=100):

    size1 = len(_global_training_data[name])
    size2 = len(_global_training_structures[name])
    kernel_vector = np.zeros(size1 * 3 + size2)

    force_vector = \
        energy_force_vector(name, energy_force_kernel, x, hyps, cutoffs,
                            hyps_mask, n_cpus, n_sample)
    energy_vector = \
        energy_energy_vector(name, energy_energy_kernel, x, hyps, cutoffs,
                             hyps_mask, n_cpus, n_sample)

    kernel_vector[0:size1*3] = force_vector
    kernel_vector[size1*3:] = energy_vector

    return kernel_vector


def efs_kern_vec(name, efs_force_kernel, efs_energy_kernel, x, hyps,
                 cutoffs=None, hyps_mask=None, n_cpus=1, n_sample=100):

    size1 = len(_global_training_data[name])
    size2 = len(_global_training_structures[name])
    tot_size = size1 * 3 + size2

    # Initialize arrays.
    energy_vector = np.zeros(tot_size)
    force_array = np.zeros((3, tot_size))
    stress_array = np.zeros((6, tot_size))

    # Compute force and energy arrays.
    force_arrays = \
        efs_force_vector(name, efs_force_kernel, x, hyps, cutoffs,
                         hyps_mask, n_cpus, n_sample)

    energy_arrays = \
        efs_energy_vector(name, efs_energy_kernel, x, hyps, cutoffs,
                          hyps_mask, n_cpus, n_sample)

    # Populate arrays.
    energy_vector[0:size1 * 3] = force_arrays[0]
    energy_vector[size1 * 3:] = energy_arrays[0]
    force_array[:, 0:size1 * 3] = force_arrays[1]
    force_array[:, size1 * 3:] = energy_arrays[1]
    stress_array[:, 0:size1 * 3] = force_arrays[2]
    stress_array[:, size1 * 3:] = energy_arrays[2]

    return energy_vector, force_array, stress_array


# --------------------------------------------------------------------------
#                      Ky hyperparameter gradient
# --------------------------------------------------------------------------

def get_ky_and_hyp_pack(name, s1, e1, s2, e2, same: bool, hyps: np.ndarray,
                        kernel_grad, cutoffs=None, hyps_mask=None):
    """
    computes a block of ky matrix and its derivative to hyper-parameter
    If the cpu set up is None, it uses as much as posible cpus

    :param hyps: list of hyper-parameters
    :param name: name of the gp instance.
    :param kernel_grad: function object of the kernel gradient
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    :return: hyp_mat, ky_mat
    """

    # assume sigma_n is the final hyperparameter
    _, non_noise_hyps, _ = obtain_noise_len(hyps, hyps_mask)

    # initialize matrices
    size1 = (e1-s1) * 3
    size2 = (e2-s2) * 3
    k_mat = np.zeros([size1, size2])
    hyp_mat = np.zeros([non_noise_hyps, size1, size2])

    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    ds = [1, 2, 3]

    training_data = _global_training_data[name]
    # calculate elements
    for m_index in range(size1):
        x_1 = training_data[int(math.floor(m_index / 3))+s1]
        d_1 = ds[m_index % 3]

        if (same):
            lowbound = m_index
        else:
            lowbound = 0
        for n_index in range(lowbound, size2):
            x_2 = training_data[int(math.floor(n_index / 3))+s2]
            d_2 = ds[n_index % 3]

            # calculate kernel and gradient
            cov = kernel_grad(x_1, x_2, d_1, d_2, *args)

            # store kernel value
            k_mat[m_index, n_index] = cov[0]
            grad = from_grad_to_mask(cov[1], hyps_mask)

            hyp_mat[:, m_index, n_index] = grad
            if (same):
                k_mat[n_index, m_index] = cov[0]
                hyp_mat[:, n_index, m_index] = grad

    return hyp_mat, k_mat


def get_ky_and_hyp(hyps: np.ndarray, name, kernel_grad, cutoffs=None,
                   hyps_mask=None, n_cpus=1, n_sample=100):
    """
    parallel version of get_ky_and_hyp

    :param hyps: list of hyper-parameters
    :param name: name of the gp instance.
    :param kernel_grad: function object of the kernel gradient
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters
    :param n_cpus: number of cpus to use.
    :param n_sample: the size of block for matrix to compute

    :return: hyp_mat, ky_mat
    """

    training_data = _global_training_data[name]
    size = len(training_data)
    size3 = size*3

    sigma_n, non_noise_hyps, train_noise = obtain_noise_len(hyps, hyps_mask)

    if (n_cpus is None):
        n_cpus = mp.cpu_count()
    if (n_cpus == 1):
        hyp_mat0, k_mat = get_ky_and_hyp_pack(
            name, 0, size, 0, size, True,
            hyps, kernel_grad, cutoffs, hyps_mask)
    else:

        block_id, nbatch = partition_matrix(n_sample, size, n_cpus)

        result_queue = mp.Queue()
        children = []
        for wid in range(nbatch):
            s1, e1, s2, e2 = block_id[wid]
            children.append(
                mp.Process(
                    target=queue_wrapper,
                    args=(result_queue, wid, get_ky_and_hyp_pack,
                          (name, s1, e1, s2, e2, s1 == s2,
                           hyps, kernel_grad, cutoffs, hyps_mask))))

        # Run child processes.
        for c in children:
            c.start()

        k_mat = np.zeros([size3, size3])
        hyp_mat0 = np.zeros([non_noise_hyps, size3, size3])
        for _ in range(nbatch):
            wid, result_chunk = result_queue.get(block=True)
            s1, e1, s2, e2 = block_id[wid]
            h_mat_block, k_mat_block = result_chunk
            k_mat[s1*3:e1*3, s2*3:e2*3] = k_mat_block
            hyp_mat0[:, s1*3:e1*3, s2*3:e2*3] = h_mat_block
            if (s1 != s2):
                k_mat[s2*3:e2*3, s1*3:e1*3] = k_mat_block.T
                for idx in range(hyp_mat0.shape[0]):
                    hyp_mat0[idx, s2*3:e2*3, s1*3:e1*3] = h_mat_block[idx].T

        # Join child processes (clean up zombies).
        for c in children:
            c.join()

    # add gradient of noise variance
    if (train_noise):
        hyp_mat = np.zeros([hyp_mat0.shape[0]+1, hyp_mat0.shape[1],
                            hyp_mat0.shape[2]])
        hyp_mat[:-1, :, :] = hyp_mat0
        hyp_mat[-1, :, :] = np.eye(size3) * 2 * sigma_n
    else:
        hyp_mat = hyp_mat0

    # matrix manipulation
    ky_mat = k_mat + sigma_n ** 2 * np.eye(size3)

    return hyp_mat, ky_mat


# --------------------------------------------------------------------------
#                        Computing the likelihood
# --------------------------------------------------------------------------

def get_like_from_mats(ky_mat, l_mat, alpha, name):
    """ compute the likelihood from the covariance matrix

    :param ky_mat: the covariance matrix

    :return: float, likelihood
    """
    # catch linear algebra errors
    force_labels = _global_training_labels[name]
    energy_labels = _global_energy_labels[name]
    labels = np.concatenate((force_labels, energy_labels))

    # calculate likelihood
    like = (-0.5 * np.matmul(labels, alpha) -
            np.sum(np.log(np.diagonal(l_mat))) -
            math.log(2 * np.pi) * ky_mat.shape[1] / 2)

    return like


def get_neg_like_grad(hyps: np.ndarray, name: str,
                      kernel_grad, logger_name: str=None,
                      cutoffs=None, hyps_mask=None,
                      n_cpus=1, n_sample=100):
    """compute the log likelihood and its gradients

    :param hyps: list of hyper-parameters
    :type hyps: np.ndarray
    :param name: name of the gp instance.
    :param kernel_grad: function object of the kernel gradient
    :param logger_name: name of logger object for dumping every hyper-parameter
                   sets computed
    :type logger_name: str
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters
    :param n_cpus: number of cpus to use.
    :param n_sample: the size of block for matrix to compute

    :return: float, np.array
    """

    time0 = time.time()

    hyp_mat, ky_mat = \
        get_ky_and_hyp(hyps, name, kernel_grad, cutoffs=cutoffs,
                       hyps_mask=hyps_mask, n_cpus=n_cpus, n_sample=n_sample)

    logger = logging.getLogger(logger_name)
    logger.debug(f"get_ky_and_hyp {time.time()-time0}")

    time0 = time.time()

    like, like_grad = \
        get_like_grad_from_mats(ky_mat, hyp_mat, name)

    logger.debug(f"get_like_grad_from_mats {time.time()-time0}")

    logger.debug('')
    logger.info(f'Hyperparameters: {list(hyps)}')
    logger.info(f'Likelihood: {like}')
    logger.info(f'Likelihood Gradient: {list(like_grad)}')

    return -like, -like_grad


def get_like_grad_from_mats(ky_mat, hyp_mat, name):
    """compute the gradient of likelihood to hyper-parameters
    from covariance matrix and its gradient

    :param ky_mat: covariance matrix
    :type ky_mat: np.array
    :param hyp_mat: dky/d(hyper parameter) matrix
    :type hyp_mat: np.array
    :param name: name of the gp instance.

    :return: float, list. the likelihood and its gradients
    """

    number_of_hyps = hyp_mat.shape[0]

    # catch linear algebra errors
    try:
        ky_mat_inv = np.linalg.inv(ky_mat)
        l_mat = np.linalg.cholesky(ky_mat)
    except np.linalg.LinAlgError:
        return -1e8, np.zeros(number_of_hyps)

    labels = _global_training_labels[name]

    alpha = np.matmul(ky_mat_inv, labels)
    alpha_mat = np.matmul(alpha.reshape(-1, 1), alpha.reshape(1, -1))
    like_mat = alpha_mat - ky_mat_inv

    # calculate likelihood
    like = (-0.5 * np.matmul(labels, alpha) -
            np.sum(np.log(np.diagonal(l_mat))) -
            math.log(2 * np.pi) * ky_mat.shape[1] / 2)

    # calculate likelihood gradient
    like_grad = np.zeros(number_of_hyps)
    for n in range(number_of_hyps):
        like_grad[n] = 0.5 * np.trace(np.matmul(like_mat, hyp_mat[n, :, :]))

    return like, like_grad



import logging
import json
import numpy as np
import time
import warnings

from copy import deepcopy
from datetime import datetime
from shutil import copyfile
from typing import List, Tuple, Union

import flare
import flare.predict as predict
from flare import struc, gp, env, md
from flare.dft_interface import dft_software
from flare.output import Output
from flare.utils.learner import is_std_in_bound
from flare.utils.element_coder import NumpyEncoder

class OTF:
    """Trains a Gaussian process force field on the fly during
        molecular dynamics.

    Args:
        dt (float): MD timestep.
        number_of_steps (int): Number of timesteps in the training
            simulation.
        prev_pos_init ([type], optional): Previous positions. Defaults
            to None.
        rescale_steps (List[int], optional): List of frames for which the
            velocities of the atoms are rescaled. Defaults to [].
        rescale_temps (List[int], optional): List of rescaled temperatures.
            Defaults to [].

        gp (gp.GaussianProcess): Initial GP model.
        calculate_energy (bool, optional): If True, the energy of each
            frame is calculated with the GP. Defaults to False.
        calculate_efs (bool, optional): If True, the energy and stress of each
            frame is calculated with the GP. Defaults to False.
        write_model (int, optional): If 0, write never. If 1, write at
            end of run. If 2, write after each training and end of run.
            If 3, write after each time atoms are added and end of run.

        std_tolerance_factor (float, optional): Threshold that determines
            when DFT is called. Specifies a multiple of the current noise
            hyperparameter. If the epistemic uncertainty on a force
            component exceeds this value, DFT is called. Defaults to 1.
        skip (int, optional): Number of frames that are skipped when
            dumping to the output file. Defaults to 0.
        init_atoms (List[int], optional): List of atoms from the input
            structure whose local environments and force components are
            used to train the initial GP model. If None is specified, all
            atoms are used to train the initial GP. Defaults to None.
        output_name (str, optional): Name of the output file. Defaults to
            'otf_run'.
        max_atoms_added (int, optional): Number of atoms added each time
            DFT is called. Defaults to 1.
        freeze_hyps (int, optional): Specifies the number of times the
            hyperparameters of the GP are optimized. After this many
            updates to the GP, the hyperparameters are frozen.
            Defaults to 10.

        force_source (Union[str, object], optional): DFT code used to calculate
            ab initio forces during training. A custom module can be used here
            in place of the DFT modules available in the FLARE package. The
            module must contain two functions: parse_dft_input, which takes a
            file name (in string format) as input and returns the positions,
            species, cell, and masses of a structure of atoms; and run_dft_par,
            which takes a number of DFT related inputs and returns the forces
            on all atoms.  Defaults to "qe".
        npool (int, optional): Number of k-point pools for DFT
            calculations. Defaults to None.
        mpi (str, optional): Determines how mpi is called. Defaults to
            "srun".
        dft_loc (str): Location of DFT executable.
        dft_input (str): Input file.
        dft_output (str): Output file.
        dft_kwargs ([type], optional): Additional arguments which are
            passed when DFT is called; keyword arguments vary based on the
            program (e.g. ESPRESSO vs. VASP). Defaults to None.
        store_dft_output (Tuple[Union[str,List[str]],str], optional):
            After DFT calculations are called, copy the file or files
            specified in the first element of the tuple to a directory
            specified as the second element of the tuple.
            Useful when DFT calculations are expensive and want to be kept
            for later use. The first element of the tuple can either be a
            single file name, or a list of several. Copied files will be
            prepended with the date and time with the format
            'Year.Month.Day:Hour:Minute:Second:'.

        n_cpus (int, optional): Number of cpus used during training.
            Defaults to 1.
    """

    def __init__(
     self,
     # md args
     dt: float, number_of_steps: int, prev_pos_init: 'ndarray' = None,
     rescale_steps: List[int] = [], rescale_temps: List[int] = [],
     # flare args
     gp: gp.GaussianProcess = None, calculate_energy: bool = False,
     calculate_efs: bool = False, write_model: int = 0,
     # otf args
     std_tolerance_factor: float = 1, skip: int = 0,
     init_atoms: List[int] = None, output_name: str = 'otf_run',
     max_atoms_added: int = 1, freeze_hyps: int = 10,
     # dft args
     force_source: str = "qe", npool: int = None, mpi: str = "srun",
     dft_loc: str = None, dft_input: str = None, dft_output='dft.out',
     dft_kwargs=None,
     store_dft_output: Tuple[Union[str, List[str]], str] = None,
     # other args
     n_cpus: int = 1, **kwargs):


        # set DFT
        self.dft_loc = dft_loc
        self.dft_input = dft_input
        self.dft_output = dft_output
        self.dft_step = True
        self.dft_count = 0
        if isinstance(force_source, str):
            self.dft_module = dft_software[force_source]
        else:
            self.dft_module = force_source

        # set md
        self.dt = dt
        self.number_of_steps = number_of_steps
        self.get_structure_from_input(prev_pos_init) # parse input file
        self.noa = self.structure.positions.shape[0]
        self.rescale_steps = rescale_steps
        self.rescale_temps = rescale_temps

        # set flare
        self.gp = gp
        # initialize local energies
        if calculate_energy:
            self.local_energies = np.zeros(self.noa)
        else:
            self.local_energies = None

        # set otf
        self.std_tolerance = std_tolerance_factor
        self.skip = skip
        self.max_atoms_added = max_atoms_added
        self.freeze_hyps = freeze_hyps
        if init_atoms is None: # set atom list for initial dft run
            self.init_atoms = [int(n) for n in range(self.noa)]
        else:
            self.init_atoms = init_atoms

        self.n_cpus = n_cpus # set number of cpus and npool for DFT runs
        self.npool = npool
        self.mpi = mpi

        self.dft_kwargs = dft_kwargs
        self.store_dft_output = store_dft_output

        # other args
        self.atom_list = list(range(self.noa))
        self.curr_step = 0

        # Set the prediction function based on user inputs.
        # Force only prediction.
        if (n_cpus > 1 and gp.per_atom_par and gp.parallel) and not \
           (calculate_energy or calculate_efs):
            self.pred_func = predict.predict_on_structure_par
        elif not (calculate_energy or calculate_efs):
            self.pred_func = predict.predict_on_structure
        # Energy and force prediction.
        elif (n_cpus > 1 and gp.per_atom_par and gp.parallel) and not \
             (calculate_efs):
            self.pred_func = predict.predict_on_structure_par_en
        elif not calculate_efs:
            self.pred_func = predict.predict_on_structure_en
        # Energy, force, and stress prediction.
        elif (n_cpus > 1 and gp.per_atom_par and gp.parallel):
            self.pred_func = predict.predict_on_structure_efs_par
        else:
            self.pred_func = predict.predict_on_structure_efs

        # set logger
        self.output = Output(output_name, always_flush=True)
        self.output_name = output_name
        self.gp_name = self.output_name + '_gp.json'
        self.checkpt_name = self.output_name + '_checkpt.json'

        self.write_model = write_model

    def run(self):
        """
        Performs an on-the-fly training run.

        If OTF has store_dft_output set, then the specified DFT files will
        be copied with the current date and time prepended in the format
        'Year.Month.Day:Hour:Minute:Second:'.
        """

        optional_dict = {"Restart": self.curr_step}
        self.output.write_header(
            str(self.gp), self.dt, self.number_of_steps, self.structure,
            self.std_tolerance, optional_dict)

        counter = 0
        self.start_time = time.time()

        while self.curr_step < self.number_of_steps:
            # run DFT and train initial model if first step and DFT is on
            if (self.curr_step == 0) and (self.std_tolerance != 0) and \
               (len(self.gp.training_data) == 0):

                # Are the recorded forces from the GP or DFT in ASE OTF?
                # When DFT is called, ASE energy, forces, and stresses should
                # get updated.
                self.initialize_train()

            # after step 1, try predicting with GP model
            else:
                # compute forces and stds with GP
                self.dft_step = False
                self.compute_properties()

                # get max uncertainty atoms
                std_in_bound, target_atoms = is_std_in_bound(
                    self.std_tolerance, self.gp.force_noise, self.structure,
                    self.max_atoms_added)

                if not std_in_bound:
                    # record GP forces
                    self.update_temperature()
                    self.record_state()
                    gp_frcs = deepcopy(self.structure.forces)

                    # run DFT and record forces
                    self.dft_step = True
                    self.run_dft()
                    dft_frcs = deepcopy(self.structure.forces)

                    # run MD step & record the state
                    self.record_state()

                    # compute mae and write to output
                    self.compute_mae(gp_frcs, dft_frcs)

                    # add max uncertainty atoms to training set
                    self.update_gp(target_atoms, dft_frcs)

            # write gp forces
            if counter >= self.skip and not self.dft_step:
                self.update_temperature()
                self.record_state()
                counter = 0

            counter += 1
            # TODO: Reinstate velocity rescaling.
            self.md_step()
            self.curr_step += 1
            self.checkpoint()

        self.output.conclude_run()

        if self.write_model >= 1:
            self.write_gp()
            self.checkpoint()

    def get_structure_from_input(self, prev_pos_init):
        positions, species, cell, masses = \
            self.dft_module.parse_dft_input(self.dft_input)

        self.structure = struc.Structure(
            cell=cell, species=species, positions=positions, mass_dict=masses,
            prev_positions=prev_pos_init, species_labels=species)

    def initialize_train(self):
        # call dft and update positions
        self.run_dft()
        dft_frcs = deepcopy(self.structure.forces)

        self.update_temperature()
        self.record_state()

        # make initial gp model and predict forces
        self.update_gp(self.init_atoms, dft_frcs)

    def compute_properties(self):
        '''
        In ASE-OTF, it will be replaced by subclass method
        '''
        self.gp.check_L_alpha()
        self.pred_func(self.structure, self.gp, self.n_cpus)

    def md_step(self):
        '''
        Take an MD step. This updates the positions of the structure.
        '''
        md.update_positions(self.dt, self.noa, self.structure)

    def write_gp(self):
        self.gp.write_model(self.gp_name)

    def run_dft(self):
        """Calculates DFT forces on atoms in the current structure.

        If OTF has store_dft_output set, then the specified DFT files will
        be copied with the current date and time prepended in the format
        'Year.Month.Day:Hour:Minute:Second:'.

        Calculates DFT forces on atoms in the current structure."""

        f = logging.getLogger(self.output.basename+'log')
        f.info('\nCalling DFT...\n')

        # calculate DFT forces
        # TODO: Return stress and energy
        forces = self.dft_module.run_dft_par(
            self.dft_input, self.structure, self.dft_loc, n_cpus=self.n_cpus,
            dft_out=self.dft_output, npool=self.npool, mpi=self.mpi,
            dft_kwargs=self.dft_kwargs)

        # Note: also need to update stresses when performing a simulation
        # in the NPT ensemble.
        self.structure.forces = forces

        # write wall time of DFT calculation
        self.dft_count += 1
        self.output.conclude_dft(self.dft_count, self.start_time)

        # Store DFT outputs in another folder if desired
        # specified in self.store_dft_output
        if self.store_dft_output is not None:
            dest = self.store_dft_output[1]
            target_files = self.store_dft_output[0]
            now = datetime.now()
            dt_string = now.strftime("%Y.%m.%d:%H:%M:%S:")
            if isinstance(target_files, str):
                to_copy = [target_files]
            else:
                to_copy = target_files
            for ofile in to_copy:
                copyfile(ofile, dest+'/'+dt_string+ofile)

    def update_gp(self, train_atoms: List[int], dft_frcs: 'ndarray',
                  dft_energy: float = None, dft_stress: 'ndarray' = None):
        """
        Updates the current GP model.


        Args:
            train_atoms (List[int]): List of atoms whose local environments
                will be added to the training set.
            dft_frcs (np.ndarray): DFT forces on all atoms in the structure.
        """
        self.output.add_atom_info(train_atoms, self.structure.stds)

        # update gp model
        self.gp.update_db(self.structure, dft_frcs, custom_range=train_atoms,
                          energy=dft_energy, stress=dft_stress)

        self.gp.set_L_alpha()

        # write model
        if (self.dft_count-1) < self.freeze_hyps:
            self.train_gp()
            if self.write_model == 2:
                self.write_gp()
        if self.write_model == 3:
            self.write_gp()

    def train_gp(self):
        """Optimizes the hyperparameters of the current GP model."""

        self.gp.train(logger_name=self.output.basename+'hyps')

        hyps, labels = self.gp.hyps_and_labels
        if labels is None:
            labels = self.gp.hyp_labels

        self.output.write_hyps(labels, hyps,
                               self.start_time,
                               self.gp.likelihood, self.gp.likelihood_gradient,
                               hyps_mask=self.gp.hyps_mask)

    def compute_mae(self, gp_frcs, dft_frcs):
        mae = np.mean(np.abs(gp_frcs - dft_frcs))
        mac = np.mean(np.abs(dft_frcs))

        f = logging.getLogger(self.output.basename+'log')
        f.info(f'mean absolute error: {mae:.4f} eV/A')
        f.info(f'mean absolute dft component: {mac:.4f} eV/A')

    def update_positions(self, new_pos: 'ndarray'):
        """Performs a Verlet update of the atomic positions.

        Args:
            new_pos (np.ndarray): Positions of atoms in the next MD frame.
        """
        if self.curr_step in self.rescale_steps:
            rescale_ind = self.rescale_steps.index(self.curr_step)
            temp_fac = self.rescale_temps[rescale_ind] / self.temperature
            vel_fac = np.sqrt(temp_fac)
            self.structure.prev_positions = \
                new_pos - self.velocities * self.dt * vel_fac
        else:
            self.structure.prev_positions = self.structure.positions
        self.structure.positions = new_pos
        self.structure.positions[:] = self.structure.wrap_positions()

    def update_temperature(self):
        """Updates the instantaneous temperatures of the system.

        Args:
            new_pos (np.ndarray): Positions of atoms in the next MD frame.
        """
        KE, temperature, velocities = \
            md.calculate_temperature(self.structure, self.dt, self.noa)
        self.KE = KE
        self.temperature = temperature
        self.velocities = velocities

    def record_state(self):
        self.output.write_md_config(
            self.dt, self.curr_step, self.structure, self.temperature,
            self.KE, self.start_time, self.dft_step, self.velocities)

    def as_dict(self):
        self.dft_module = self.dft_module.__name__
        out_dict = deepcopy(dict(vars(self)))
        self.dft_module = eval(self.dft_module)

        out_dict['gp'] = self.gp_name
        out_dict['structure'] = self.structure.as_dict()

        for key in ['output', 'pred_func']:
            out_dict.pop(key)

        return out_dict

    @staticmethod
    def from_dict(in_dict):
        if in_dict['write_model'] <= 1: # TODO: detect GP version
            warnings.warn("The GP model might not be the latest")
       
        gp_model = gp.GaussianProcess.from_file(in_dict['gp'])
        in_dict['gp'] = gp_model
        in_dict['structure'] = struc.Structure.from_dict(in_dict['structure'])

        if 'flare.dft_interface' in in_dict['dft_module']:
            for dft_name in ['qe', 'cp2k', 'vasp']:
                if dft_name in in_dict['dft_module']:
                    in_dict['force_source'] = dft_name
                    break
        else: # if force source is a module
            in_dict['force_source'] = eval(in_dict['dft_module'])

        new_otf = OTF(**in_dict)
        new_otf.structure = in_dict['structure']
        new_otf.dft_count = in_dict['dft_count']
        new_otf.curr_step = in_dict['curr_step']
        return new_otf

    def checkpoint(self):
        name = self.checkpt_name
        if '.json' != name[-5:]:
            name += '.json'
        with open(name, 'w') as f:
            json.dump(self.as_dict(), f, cls=NumpyEncoder)

    @classmethod
    def from_checkpoint(cls, filename):
        with open(filename, 'r') as f:
            otf_model = cls.from_dict(json.loads(f.readline()))

        return otf_model



"""
The :class:`Structure` object is a collection of atoms in a periodic box.
The mandatory inputs are the cell vectors of the box and the chemical species
and *Cartesian coordinates* of the atoms.
The atoms are automatically folded back into the primary cell, so the
input coordinates don't need to lie inside the box.
"""
import numpy as np
from flare.utils.element_coder import element_to_Z, Z_to_element, NumpyEncoder
from flare.utils.learner import get_max_cutoff
from json import dumps, loads

from typing import List, Union, Any

try:
    # Used for to_pmg_structure method
    import pymatgen.core.structure as pmgstruc
    import pymatgen.io.vasp.inputs as pmgvaspio
    _pmg_present = True
except ImportError:
    _pmg_present = False


class Structure:
    """
    Contains information about a periodic structure of atoms, including the
    periodic cell boundaries, atomic species, and coordinates.

    *Note that input positions are assumed to be Cartesian.*

    :param cell: 3x3 array whose rows are the Bravais lattice vectors of the
        cell.
    :type cell: np.ndarray
    :param species: List of atomic species, which are represented either as
        integers or chemical symbols.
    :type species: List
    :param positions: Nx3 array of atomic coordinates.
    :type positions: np.ndarray
    :param mass_dict: Dictionary of atomic masses used in MD simulations.
    :type mass_dict: dict
    :param prev_positions: Nx3 array of previous atomic coordinates used in
        MD simulations.
    :type prev_positions: np.ndarray
    :param species_labels: List of chemical symbols. Used in the output file
        of on-the-fly runs.
    :type species_labels: List[str]
    :param stds: Uncertainty associated with forces
    :type stds: np.ndarray
    """

    def __init__(self, cell: 'ndarray', species: Union[List[str], List[int]],
                 positions: 'ndarray', mass_dict: dict = None,
                 prev_positions: 'ndarray' = None,
                 species_labels: List[str] = None,
                 forces=None, stds=None, energy: float=None):

        # Define cell (each row is a Bravais lattice vector).
        self.cell = np.array(cell)

        # Compute the max cutoff compatible with a 3x3x3 supercell of the
        # structure.
        self.max_cutoff = get_max_cutoff(self.cell)

        # Set positions.
        self.positions = np.array(positions)

        # If species are strings, convert species to integers by atomic number
        if species_labels is None:
            self.species_labels = species
        else:
            self.species_labels = species_labels
        self.coded_species = np.array([element_to_Z(spec) for spec in species])
        self.nat = len(species)

        # Default: atoms have no velocity
        if prev_positions is None:
            self.prev_positions = np.copy(self.positions)
        else:
            assert len(positions) == len(prev_positions), \
                'Previous positions and positions are not same length'
            self.prev_positions = prev_positions

        # Set forces, energies, and stresses and their uncertainties.
        if forces is not None:
            self.forces = np.array(forces)
        else:
            self.forces = np.zeros((len(positions), 3))

        if stds is not None:
            self.stds = np.array(stds)
        else:
            self.stds = np.zeros((len(positions), 3))

        self.energy = energy

        self.local_energies = None
        self.local_energy_stds = None
        self.partial_stresses = None
        self.partial_stress_stds = None
        self.stress = None
        self.stress_stds = None
        self.potential_energy = None # duplicated with self.energy?

        self.mass_dict = mass_dict

        # Convert from elements to atomic numbers in mass dict
        if mass_dict is not None:
            keys = list(mass_dict.keys())
            for elt in keys:
                if isinstance(elt, str):
                    mass_dict[element_to_Z(elt)] = mass_dict[elt]
                    if elt.isnumeric():
                        mass_dict[int(elt)] = mass_dict[elt]

    @property
    def positions(self):
        return self._positions

    @property
    def wrapped_positions(self):
        return self._wrapped_positions

    @positions.setter
    def positions(self, position_array):
        self._positions = position_array
        self._wrapped_positions = self.wrap_positions()

    @property
    def cell(self):
        return self._cell

    @property
    def vec1(self):
        return self._vec1

    @property
    def vec2(self):
        return self._vec2

    @property
    def vec3(self):
        return self._vec3

    @property
    def cell_transpose(self):
        return self._cell_transpose

    @property
    def cell_transpose_inverse(self):
        return self._cell_transpose_inverse

    @property
    def cell_dot(self):
        return self._cell_dot

    @property
    def cell_dot_inverse(self):
        return self._cell_dot_inverse

    @cell.setter
    def cell(self, cell_array):
        """Set the cell and related properties."""
        self._cell = cell_array
        self._vec1 = cell_array[0, :]
        self._vec2 = cell_array[1, :]
        self._vec3 = cell_array[2, :]
        self._cell_transpose = cell_array.transpose()
        self._cell_transpose_inverse = np.linalg.inv(self._cell_transpose)
        self._cell_dot = self.get_cell_dot(cell_array)
        self._cell_dot_inverse = np.linalg.inv(self._cell_dot)

    @staticmethod
    def get_cell_dot(cell_array):
        """
        Compute 3x3 array of dot products of cell vectors used to
        fold atoms back to the unit cell.

        :return: 3x3 array of cell vector dot products.
        :rtype: np.ndarray
        """

        cell_dot = np.zeros((3, 3))

        for m in range(3):
            for n in range(3):
                cell_dot[m, n] = np.dot(cell_array[m], cell_array[n])

        return cell_dot

    @staticmethod
    def raw_to_relative(positions: 'ndarray', cell_transpose: 'ndarray',
                        cell_dot_inverse: 'ndarray') -> 'ndarray':
        """Convert Cartesian coordinates to relative (fractional) coordinates,
        expressed in terms of the cell vectors set in self.cell.

        :param positions: Cartesian coordinates.
        :type positions: np.ndarray
        :param cell_transpose: Transpose of the cell array.
        :type cell_transpose: np.ndarray
        :param cell_dot_inverse: Inverse of the array of dot products of
            cell vectors.
        :type cell_dot_inverse: np.ndarray
        :return: Relative positions.
        :rtype: np.ndarray
        """

        relative_positions = \
            np.matmul(np.matmul(positions, cell_transpose),
                      cell_dot_inverse)

        return relative_positions

    @staticmethod
    def relative_to_raw(relative_positions: 'ndarray',
                        cell_transpose_inverse: 'ndarray',
                        cell_dot: 'ndarray') -> 'ndarray':
        """Convert fractional coordinates to raw (Cartesian) coordinates.

        :param relative_positions: fractional coordinates.
        :type relative_positions: np.ndarray
        :param cell_transpose_inverse: Transpose of the cell array.
        :type cell_transpose_inverse: np.ndarray
        :param cell_dot: Dot products of cell vectors
        :type cell_dot: np.ndarray
        :return: Cartesian positions.
        :rtype: np.ndarray
        """

        return np.matmul(np.matmul(relative_positions, cell_dot),
                         cell_transpose_inverse)

    def wrap_positions(self) -> 'ndarray':
        """
        Convenience function which folds atoms outside of the unit cell back
        into the unit cell. in_place flag controls if the wrapped positions
        are set in the class.

        :return: Cartesian coordinates of positions all in unit cell
        :rtype: np.ndarray
        """
        rel_pos = \
            self.raw_to_relative(self.positions, self.cell_transpose,
                                 self.cell_dot_inverse)

        rel_wrap = rel_pos - np.floor(rel_pos)

        pos_wrap = self.relative_to_raw(rel_wrap, self.cell_transpose_inverse,
                                        self.cell_dot)

        return pos_wrap

    def indices_of_specie(self, specie: Union[int, str]) -> List[int]:
        """
        Return the indices of a given species within atoms of the structure.

        :param specie: Element to target, can be string or integer
        :return: The indices in the structure at which this element occurs
        :rtype: List[str]
        """
        return [i for i, spec in enumerate(self.coded_species)
                if spec == specie]

    # TODO make more descriptive
    def __str__(self) -> str:
        """
        Simple descriptive string of structure.

        :return: One-line descriptor of number of atoms and species present.
        :rtype: str
        """

        return 'Structure with {} atoms of types {}'\
            .format(self.nat, set(self.species_labels))

    def __len__(self) -> int:
        """
        Returns number of atoms in structure.

        :return: number of atoms in structure.
        :rtype: int
        """
        return self.nat

    def as_dict(self) -> dict:
        """
        Returns structure as a dictionary; useful for serialization purposes.

        :return: Dictionary version of current structure
        :rtype: dict
        """
        return dict(vars(self))

    def as_str(self) -> str:
        """
        Returns string dictionary serialization cast as string.

        :return: output of as_dict method cast as string
        :rtype: str
        """
        return dumps(self.as_dict(), cls=NumpyEncoder)

    @staticmethod
    def from_dict(dictionary: dict) -> 'flare.struc.Structure':
        """
        Assembles a Structure object from a dictionary parameterizing one.

        :param dictionary: dict describing structure parameters.
        :return: FLARE structure assembled from dictionary
        """
        struc = Structure(cell=np.array(dictionary.get('_cell',
                                                       dictionary.get(
                                                           'cell'))),
                          positions=np.array(dictionary.get('_positions',
                                                       dictionary.get(
                                                           'positions'))),
                          species=dictionary['coded_species'],
                          forces=np.array(dictionary.get('forces')),
                          mass_dict=dictionary.get('mass_dict'),
                          species_labels=dictionary.get('species_labels'),
                          energy=dictionary.get('energy', None))

        struc.stds = np.array(dictionary.get('stds'))

        return struc

    @staticmethod
    def from_ase_atoms(atoms: 'ase.Atoms', cell=None) -> 'flare.struc.Structure':
        """
        From an ASE Atoms object, return a FLARE structure

        :param atoms: ASE Atoms object
        :type atoms: ASE Atoms object
        :return: A FLARE structure from an ASE atoms object
        """

        if cell is None:
            cell = np.array(atoms.cell)

        try:
            forces = atoms.get_forces()
        except:
            forces = None 

        try:
            stds = atoms.get_uncertainties()
        except:
            stds = None

        try:
            energy = atoms.get_potential_energy()
        except:
            energy = None

        try:
            stress = atoms.get_stress()
        except:
            stress = None 

        struc = Structure(
            cell = cell, 
            positions = atoms.positions,
            species = atoms.get_chemical_symbols(),
            forces = forces,
            stds = stds,
            energy = energy,
        )
        struc.stress = stress
        return struc

    def to_ase_atoms(self) -> 'ase.Atoms':
        from ase import Atoms
        from ase.calculators.singlepoint import SinglePointCalculator

        atoms =  Atoms(self.species_labels, positions=self.positions,
                       cell=self.cell, pbc=True)

        results = {}
        properties = ["forces", "energy", "stress"]
        for p in properties:
            results[p] = getattr(self, p)
        calculator = SinglePointCalculator(atoms, **results)
        atoms.set_calculator(calculator)

        return atoms

    def to_pmg_structure(self):
        """
        Returns FLARE structure as a pymatgen structure.

        :return: Pymatgen structure corresponding to current FLARE structure
        """

        if not _pmg_present:
            raise ModuleNotFoundError("Pymatgen is not present. Please "
                                      "install Pymatgen and try again")

        if self.forces is None:
            forces_temp = np.zeros((len(self.positions), 3))
            site_properties = {'force:': forces_temp, 'std': self.stds}
        else:
            site_properties = {'force:': self.forces, 'std': self.stds}

        return pmgstruc.Structure(lattice=self.cell,
                                  species=self.species_labels,
                                  coords=self.positions,
                                  coords_are_cartesian=True,
                                  site_properties=site_properties
                                  )

    @staticmethod
    def from_pmg_structure(structure: 'pymatgen Structure') -> \
            'flare Structure':
        """
        Returns Pymatgen structure as FLARE structure.

        :param structure: Pymatgen Structure
        :type structure: Pymatgen Structure
        :return: FLARE Structure
        """

        cell = structure.lattice.matrix.copy()
        species = [str(spec) for spec in structure.species]
        positions = structure.cart_coords.copy()

        new_struc = Structure(cell=cell, species=species,
                              positions=positions)

        site_props = structure.site_properties

        if 'force' in site_props.keys():
            forces = site_props['force']
            new_struc.forces = [np.array(force) for force in forces]

        if 'std' in site_props.keys():
            stds = site_props['std']
            new_struc.stds = [np.array(std) for std in stds]

        return new_struc

    def is_valid(self, tolerance: float = .5)->bool:
        """
        Plugin to pymatgen's is_valid method to gauge if a structure
        has atoms packed too closely together, which is likely
        to cause unphysically large forces / energies or present convergence
        issues in DFT.
        :return:
        """
        pmg_structure = self.to_pmg_structure()
        return pmg_structure.is_valid(tol=tolerance)


    def to_xyz(self, extended_xyz: bool = True, print_stds: bool = False,
               print_forces: bool = False, print_max_stds: bool = False,
               print_energies: bool = False, predict_energy = None,
               dft_forces = None, dft_energy = None, timestep=-1,
               write_file: str = '', append: bool = False) -> str:
        """
        Convenience function which turns a structure into an extended .xyz
        file; useful for further input into visualization programs like VESTA
        or Ovito. Can be saved to an output file via write_file.

        :param print_stds: Print the stds associated with the structure.
        :param print_forces:
        :param extended_xyz:
        :param print_max_stds:
        :param write_file:
        :return:
        """
        species_list = [Z_to_element(x) for x in self.coded_species]
        xyz_str = ''
        xyz_str += f'{len(self.coded_species)} \n'

        # Add header line with info about lattice and properties if extended
        #  xyz option is called.
        if extended_xyz:
            cell = self.cell

            xyz_str += f'Lattice="{cell[0,0]} {cell[0,1]} {cell[0,2]}'
            xyz_str += f' {cell[1,0]} {cell[1,1]} {cell[1,2]}'
            xyz_str += f' {cell[2,0]} {cell[2,1]} {cell[2,2]}"'
            if timestep > 0:
                xyz_str += f' Timestep={timestep}'
            if predict_energy:
                xyz_str += f' PE={predict_energy}'
            if dft_energy is not None:
                xyz_str += f' DFT_PE={dft_energy}'
            xyz_str += f' Proprties="species:S:1:pos:R:3'

            if print_stds:
                xyz_str += ':stds:R:3'
                stds = self.stds
            if print_forces:
                xyz_str += ':forces:R:3'
                forces = self.forces
            if print_max_stds:
                xyz_str += ':max_std:R:1'
                stds = self.stds
            if print_energies:
                if self.local_energies is None:
                    print_energies = False
                else:
                    xyz_str += ':local_energy:R:1'
                    local_energies = self.local_energies
            if dft_forces is not None:
                xyz_str += ':dft_forces:R:3'
            xyz_str += '\n'
        else:
            xyz_str += '\n'

        for i, pos in enumerate(self.positions):
            # Write positions
            xyz_str += f"{species_list[i]} {pos[0]} {pos[1]} {pos[2]}"

            # If extended XYZ: Add in extra information
            if print_stds and extended_xyz:
                xyz_str += f" {stds[i,0]} {stds[i,1]} {stds[i,2]}"
            if print_forces and extended_xyz:
                xyz_str += f" {forces[i,0]} {forces[i,1]} {forces[i,2]}"
            if print_energies and extended_xyz:
                xyz_str += f" {local_energies[i]}"
            if print_max_stds and extended_xyz:
                xyz_str += f" {np.max(stds[i,:])} "
            if dft_forces is not None:
                xyz_str += f' {dft_forces[i, 0]} {dft_forces[i,1]} ' \
                          f'{dft_forces[i, 2]}'
            if i < (len(self.positions)-1):
                xyz_str += '\n'

        # Write to file, optionally
        if write_file:
            if append:
                fmt = 'a'
            else:
                fmt = 'w'
            with open(write_file, fmt) as f:
                f.write(xyz_str)
                f.write("\n")

        return xyz_str

    @staticmethod
    def from_file(file_name, format='') -> Union['flare.struc.Structure',
                                                 List['flare.struc.Structure']
                                                 ]:
        """
        Load a FLARE structure from a file or a series of FLARE structures
        :param file_name:
        :param format:
        :return:
        """

        # Ensure the file specified exists.
        with open(file_name, 'r') as _:
            pass

        if 'xyz' in file_name or 'xyz' in format.lower():
            raise NotImplementedError

        if 'json' in format.lower() or '.json' in file_name:
            # Assumed format is one FLARE structure per line,
            # or one line with many FLARE structures
            with open(file_name, 'r') as f:
                thelines = f.readlines()

                non_empty_lines = [loads(line) for line in thelines if
                                   len(line) > 2]

            structures = [Structure.from_dict(struc_dict) for struc_dict in
                          non_empty_lines]

            if len(structures) == 1:
                return structures[0]
            else:
                return structures

        is_poscar = 'POSCAR' in file_name or 'CONTCAR' in file_name \
            or 'vasp' in format.lower()
        if is_poscar and _pmg_present:
            pmg_structure = pmgvaspio.Poscar.from_file(file_name).structure
            return Structure.from_pmg_structure(pmg_structure)
        elif is_poscar and not _pmg_present:
            raise ImportError("Pymatgen not imported; "
                              "functionality requires pymatgen.")


def get_unique_species(species: List[Any]) -> (List, List[int]):
    """
    Returns a list of the unique species passed in, and a list of
    integers indexing them.

    :param species: Species to index uniquely
    :return: List of the unique species, and integer indexes
    """
    unique_species = []
    coded_species = []
    for spec in species:
        if spec in unique_species:
            coded_species.append(unique_species.index(spec))
        else:
            coded_species.append(len(unique_species))
            unique_species.append(spec)
    coded_species = np.array(coded_species)

    return unique_species, coded_species



"""
Utility functions for various tasks.
"""
from warnings import warn
from json import JSONEncoder
from typing import List
from math import inf

import numpy as np


def get_random_velocities(noa: int, temperature: float, mass: float):
    """Draw velocities from the Maxwell-Boltzmann distribution, assuming a
    fixed mass for all particles in amu.

    Args:
        noa (int): Number of atoms in the system.
        temperature (float): Temperature of the system.
        mass (float): Mass of each particle in amu.

    Returns:
        np.ndarray: Particle velocities, corrected to give zero center of mass motion.
    """

    # Use FLARE mass units (time = ps, length = A, energy = eV)
    mass_md = mass * 0.000103642695727
    kb = 0.0000861733034
    std = np.sqrt(kb * temperature / mass_md)
    velocities = np.random.normal(scale=std, size=(noa, 3))

    # Remove center-of-mass motion
    vel_sum = np.sum(velocities, axis=0)
    corrected_velocities = velocities - vel_sum / noa

    return corrected_velocities


def multicomponent_velocities(temperature: float, masses: List[float]):
    """Draw velocities from the Maxwell-Boltzmann distribution for particles of
    varying mass.

    Args:
        temperature (float): Temperature of the system.
        masses (List[float]): Particle masses in amu.

    Returns:
        np.ndarray: Particle velocities, corrected to give zero center of mass motion.
    """

    noa = len(masses)
    velocities = np.zeros((noa, 3))
    kb = 0.0000861733034
    mom_tot = np.array([0., 0., 0.])

    for n, mass in enumerate(masses):
        # Convert to FLARE mass units (time = ps, length = A, energy = eV)
        mass_md = mass * 0.000103642695727
        std = np.sqrt(kb * temperature / mass_md)
        rand_vel = np.random.normal(scale=std, size=(3))
        velocities[n] = rand_vel
        mom_curr = rand_vel * mass_md
        mom_tot += mom_curr

    # Correct momentum, remove center of mass motion
    mom_corr = mom_tot / noa
    for n, mass in enumerate(masses):
        mass_md = mass * 0.000103642695727
        velocities[n] -= mom_corr / mass_md

    return velocities


def get_supercell_positions(sc_size: int, cell: np.ndarray,
                            positions: np.ndarray):
    """Returns the positions of a supercell of atoms, with the number of cells
    in each direction fixed.

    Args:
        sc_size (int): Size of the supercell.
        cell (np.ndarray): 3x3 array of cell vectors.
        positions (np.ndarray): Positions of atoms in the unit cell.

    Returns:
        np.ndarray: Positions of atoms in the supercell.
    """

    sc_positions = []
    for m in range(sc_size):
        vec1 = m * cell[0]
        for n in range(sc_size):
            vec2 = n * cell[1]
            for p in range(sc_size):
                vec3 = p * cell[2]

                # append translated positions
                for pos in positions:
                    sc_positions.append(pos+vec1+vec2+vec3)

    return np.array(sc_positions)


def supercell_custom(cell: np.ndarray, positions: np.ndarray,
                     size1: int, size2: int, size3: int):
    """Returns the positions of a supercell of atoms with a chosen number of
    cells in each direction.

    Args:
        cell (np.ndarray): 3x3 array of cell vectors.
        positions (np.ndarray): Positions of atoms in the unit cell.
        size1 (int): Number of cells along the first cell vector.
        size2 (int): Number of cells along the second cell vector.
        size3 (int): Number of cells along the third cell vector.

    Returns:
        np.ndarray: Positions of atoms in the supercell.
    """

    sc_positions = []
    for m in range(size1):
        vec1 = m * cell[0]
        for n in range(size2):
            vec2 = n * cell[1]
            for p in range(size3):
                vec3 = p * cell[2]

                # append translated positions
                for pos in positions:
                    sc_positions.append(pos+vec1+vec2+vec3)

    return np.array(sc_positions)



"""
For multi-component systems, the configurational space can be highly complicated.
One may want to use different hyper-parameters and cutoffs for different interactions,
or do constraint optimisation for hyper-parameters.

To use more hyper-parameters, we need special kernel function that can differentiate different
pairs, triplets and other descriptors and determine which number to use for what interaction.

This kernel can be enabled by using the ``hyps_mask`` argument of the GaussianProcess class.
It contains multiple arrays to describe how to break down the array of hyper-parameters and
apply them when computing the kernel. Detail descriptions of this argument can be seen in
kernel/mc_sephyps.py.

The ParameterHelper class is to generate the hyps_mask with a more human readable interface.

Example:

>>> pm = ParameterHelper(species=['C', 'H', 'O'],
...                      kernels={'twobody':[['*', '*'], ['O','O']],
...                      'threebody':[['*', '*', '*'],
...                          ['O','O', 'O']]},
...                      parameters={'twobody0':[1, 0.5, 1], 'twobody1':[2, 0.2, 2],
...                            'threebody0':[1, 0.5], 'threebody1':[2, 0.2],
...                            'cutoff_threebody':1},
...                      constraints={'twobody0':[False, True]})
>>> hm = pm.as_dict()
>>> kernels = hm['kernels']
>>> gp_model = GaussianProcess(kernels=kernels,
...                            hyps=hyps, hyps_mask=hm)

In this example, four atomic species are involved. There are many kinds
of twobodys and threebodys. But we only want to use eight different signal variance
and length-scales.

In order to do so, we first define all the twobodys to be group "twobody0", by
listing "*-*" as the first element in the twobody argument. The second
element O-O is then defined to be group "twobody1". Note that the order
matters here. The later element overrides the ealier one. If
twobodys=[['O', 'O'], ['*', '*']], then all twobodys belong to group "twobody1".

Similarly, O-O-O is defined as threebody1, while all remaining ones
are left as threebody0.

The hyperpameters for each group is listed in the order of
[sig, ls, cutoff] in the parameters argument.  So in this example,
O-O interaction will use [2, 0.2, 2] as its sigma, length scale, and
cutoff.

For threebody, the parameter arrays only come with two elements. So there
is no cutoff associated with threebody0 or threebody1; instead, a universal
cutoff is used, which is defined as 'cutoff_threebody'.

The constraints argument define which hyper-parameters will be optimized.
True for optimized and false for being fixed.

Here are a couple more simple examples.

Define a 5-parameter 2+3 kernel (1, 0.5, 1, 0.5, 0.05)

>>> pm = ParameterHelper(kernels=['twobody', 'threebody'],
...                     parameters={'sigma': 1,
...                                 'lengthscale': 0.5,
...                                 'cutoff_twobody': 2,
...                                 'cutoff_threebody': 1,
...                                 'noise': 0.05})

Define a 5-parameter 2+3 kernel (1, 1, 1, 1, 0.05)

>>> pm = ParameterHelper(kernels=['twobody', 'threebody'],
...                      parameters={'cutoff_twobody': 2,
...                                  'cutoff_threebody': 1,
...                                  'noise': 0.05},
...                      ones=ones,
...                      random=not ones)

Define a 9-parameter 2+3 kernel

>>> pm = ParameterHelper()
>>> pm.define_group('specie', 'O', ['O'])
>>> pm.define_group('specie', 'rest', ['C', 'H'])
>>> pm.define_group('twobody', '**', ['*', '*'])
>>> pm.define_group('twobody', 'OO', ['O', 'O'])
>>> pm.define_group('threebody', '***', ['*', '*', '*'])
>>> pm.define_group('threebody', 'Oall', ['O', 'O', 'O'])
>>> pm.set_parameters('**', [1, 0.5])
>>> pm.set_parameters('OO', [1, 0.5])
>>> pm.set_parameters('Oall', [1, 0.5])
>>> pm.set_parameters('***', [1, 0.5])
>>> pm.set_parameters('cutoff_twobody', 5)
>>> pm.set_parameters('cutoff_threebody', 4)

See more examples in functions ``ParameterHelper.define_group`` , ``ParameterHelper.set_parameters``,
and in the tests ``tests/test_parameters.py``

If you want to add in a new hyperparameter set to an already-existing GP, you can perform the
following steps:

>> hyps_mask = pm.as_dict()
>> hyps = hyps_mask['hyps']
>> kernels = hyps_mask['kernels']
>> gp_model.update_kernel(kernels, 'mc', hyps_mask)
>> gp_model.hyps = hyps
"""

import inspect
import json
import logging
import math
import numpy as np
import pickle
import time

from copy import deepcopy
from itertools import combinations_with_replacement, permutations
from numpy import array as nparray
from numpy import max as npmax
from numpy.random import random as nprandom
from typing import List, Callable, Union

from flare.output import set_logger
from flare.parameters import Parameters
from flare.utils.element_coder import element_to_Z, Z_to_element


class ParameterHelper():
    """
    A helper class to construct the hyps_mask dictionary for AtomicEnvironment
    , GaussianProcess and MappedGaussianProcess

    Args:
        hyps_mask (dict): Not implemented yet
        species (dict, list): Define specie groups
        kernels (dict, list): Define kernels and groups for the kernels
        cutoff_groups (dict): Define different cutoffs for different species
        parameters (dict): Define signal variance, length scales, and cutoffs
        constraints (dict): If listed as False, the cooresponding hyperparmeters
            will not be trained
        allseparate (bool): If True, define each type pair/triplet into a
            separate group.
        random (bool): If True, randomized all signal variances and lengthscales
        one (bool): If True, set all signal variances and lengthscales to one
        verbose (str): Level to print with "ERROR", "WARNING", "INFO", "DEBUG"

    * the ``species`` is an optional input. It can be left as None if the user only wants
      to set up one group of hyper-parameters for each kernel.
    * the ``kernels`` can be defined along with or without groups. But the later mode
      is not compatible with the ``allseparate`` flag.

      >>> kernels=['twobody', 'threebody'],

      or

      >>> kernels={'twobody':[['*', '*'], ['O','O']],
      ...          'threebody':[['*', '*', '*'],
      ...                       ['O','O', 'O']]},

      Current options for the kernels are twobody, threebody and manybody (based on coordination number).
    * See format of ``species``, ``kernels`` (dict), and ``cutoff_groups`` in ``list_groups()`` function.
    * See format of ``parameters`` and ``constraints`` in ``list_parameters()`` function.
    """

    # TO DO, sync it to kernel class
    #  need to be synced with kernel class

    # name of the kernels
    all_kernel_types = ['twobody', 'threebody', 'manybody']
    cutoff_types = {'cut3b': 'threebody'}
    cutoff_types_keys = list(cutoff_types.keys())
    cutoff_types_values = list(cutoff_types.values())
    additional_groups = []

    # dimension of the kernels
    ndim = {'twobody': 2, 'threebody': 3, 'manybody': 2, 'cut3b': 2}
    n_kernel_parameters = {'twobody': 2,
                           'threebody': 2, 'manybody': 2, 'cut3b': 0}

    def __init__(self, hyps_mask=None, species=None, kernels={},
                 cutoff_groups={}, parameters=None,
                 constraints={}, allseparate=False, random=False, ones=False,
                 verbose="WARNING"):

        self.logger = set_logger("ParameterHelper", stream=True,
                                 fileout_name=None, verbose=verbose)

        self.all_group_types = ParameterHelper.all_kernel_types + \
            self.cutoff_types_keys + \
            self.additional_groups

        self.all_types = ['specie'] + self.all_group_types

        # number of groups {'twobody': 1, 'threebody': 2}
        self.n = {}
        # definition of groups {'specie': [['C', 'H'], ['O']],
        # 'twobody': [[['*', '*']], [[ele1, ele2]]]}
        self.groups = {}
        # joint values of the groups {'specie': ['C', 'H', 'O'],
        # 'twobody': [['*', '*'], [ele1, ele2]]}
        self.all_members = {}
        # names of each group {'specie': ['group1', 'group2'], 'twobody':
        # ['twobody0', 'twobody1']}
        self.all_group_names = {}
        # joint list of all the keys in self.all_group_names
        self.all_names = []

        # set up empty container
        for group_type in self.all_types:
            self.n[group_type] = 0
            self.groups[group_type] = []
            self.all_members[group_type] = []
            self.all_group_names[group_type] = []

        # store parameters, key should be the one used in
        # all_group_names or kernel_name
        self.sigma = {}
        self.ls = {}
        self.noise = 0.05
        self.energy_noise = 0.1
        self.opt = {'noise': True}

        # key should be sigma, lengthscale
        # cutoff_kernel_name
        self.universal = {}

        # key should be in all_group_names
        self.all_cutoff = {}

        # used for as_dict
        self.hyps_sig = {}
        self.hyps_ls = {}
        self.hyps_opt = {}
        self.cutoff_list = {}
        self.mask = {}

        self.hyps = None

        if isinstance(kernels, dict):
            self.kernel_dict = kernels
            self.kernels = list(kernels.keys())
            assert (not allseparate)
        elif isinstance(kernels, list):
            self.kernels = kernels
            # by default, there is only one group of hyperparameters
            # for each type of the kernel
            # unless allseparate is defined
            self.kernel_dict = {}
            for ktype in kernels:
                self.kernel_dict[ktype] = [['*']*ParameterHelper.ndim[ktype]]

        if species is not None:
            self.list_groups('specie', species)

            # define groups
            if allseparate:
                for ktype in self.kernels:
                    self.all_separate_groups(ktype)
            else:
                for ktype in self.kernels:
                    self.list_groups(ktype, self.kernel_dict[ktype])

            # check for cut3b
            for group in cutoff_groups:
                self.list_groups(group, cutoff_groups[group])

            # define parameters
            if parameters is not None:
                self.list_parameters(parameters, constraints)

            if 'lengthscale' in self.universal or 'sigma' in self.universal:
                universal = True
            else:
                universal = False

            if (random+ones+universal) > 1:
                raise RuntimeError(
                    "random and ones cannot be simultaneously True")
            elif random or ones or universal:
                for ktype in self.kernels:
                    self.fill_in_parameters(
                        ktype, random=random, ones=ones, universal=universal)

        elif len(self.kernels) > 0:
            self.list_groups('specie', ['*'])

            # define groups
            for ktype in self.kernels:
                self.list_groups(ktype, self.kernel_dict[ktype])

            # check for cut3b
            for group in cutoff_groups:
                self.list_groups(group, cutoff_groups[group])

            # define parameters
            if parameters is not None:
                self.list_parameters(parameters, constraints)

            if 'lengthscale' in self.universal or 'sigma' in self.universal:
                universal = True
            else:
                universal = False

            if (random+ones+universal) > 1:
                raise RuntimeError(
                    "random and ones cannot be simultaneously True")
            elif random or ones or universal:
                for ktype in self.kernels:
                    self.fill_in_parameters(
                        ktype, random=random, ones=ones, universal=universal)

    def list_parameters(self, parameter_dict: dict, constraints: dict = {}):
        """Define many groups of parameters

        Args:
            parameter_dict (dict): dictionary of all parameters
            constraints (dict): dictionary of all constraints

        Example:

        >>> parameter_dict={"group_name":[sig, ls, cutoffs], ...}
        >>> constraints={"group_name":[True, False, False], ...}

        The name of parameters can be the group name previously defined in
        define_group or list_groups function. Aside from the group name,
        ``noise``, ``cutoff_twobody``, ``cutoff_threebody``, and
        ``cutoff_manybody`` are reserved for noise parmater
        and universal cutoffs, while ``sigma`` and ``lengthscale`` are
        reserved for universal signal variances and length scales.

        For non-reserved keys, the value should be a list of 2 to 3 elements,
        corresponding to the sigma, lengthscale and cutoff (if the third one
        is defined). For reserved keys, the value should be a float number.

        The parameter_dict and constraints should use the same set of keys.
        If a key in constraints is not used in parameter_dict, it will be ignored.

        The value in the constraints can be either a single bool, which apply
        to all parameters, or list of bools that apply to each parameter.
        """

        for name in parameter_dict:
            self.set_parameters(
                name, parameter_dict[name], constraints.get(name, True))
        for name in constraints:
            if name not in parameter_dict:
                self.set_constraints(
                    name, constraints[name])

    def list_groups(self, group_type, definition_list):
        """define groups in batches.

        Args:
            group_type (str): "specie", "twobody", "threebody", "cut3b", "manybody"
            definition_list (list, dict): list of elements

        This function runs define_group in batch. Please first read
        the manual of define_group.

        If the definition_list is a list, it is equivalent to
        executing define_group through the definition_list.

        >>> for all terms in the list:
        >>>     define_group(group_type, group_type+'n', the nth term in the list)

        So the first twobody defined will be group twobody0, second one will be
        group twobody1. For specie, it will define all the listed elements as
        groups with only one element with their original name.

        If the definition_list is a dictionary, it is equivalent to

        >>> for k, v in the dict:
        >>>     define_group(group_type, k, v)

        It is not recommended to use the dictionary mode, especially when
        the group definitions are conflicting with each other. There is no
        guarantee that the priority order is the same as you want.

        Unlike ParameterHelper.define_group(), it can only be called once for each
        group_type, and not after any ParameterHelper.define_group() calls.

        """

        if group_type == 'specie':
            if len(self.all_group_names['specie']) > 0:
                raise RuntimeError("this function has to be run "
                                   "before any define_group")
            if isinstance(definition_list, list):
                for ele in definition_list:
                    if isinstance(ele, list):
                        self.define_group('specie', ele, ele)
                    else:
                        self.define_group('specie', ele, [ele])
            elif isinstance(definition_list, dict):
                for ele in definition_list:
                    self.define_group('specie', ele, definition_list[ele])
            else:
                raise RuntimeError("type unknown")
        else:
            if self.n['specie'] == 0:
                raise RuntimeError("this function has to be run "
                                   "before any define_group")
            if isinstance(definition_list, list):
                ngroup = len(definition_list)
                for idg in range(ngroup):
                    self.define_group(group_type, f"{group_type}{idg}",
                                      definition_list[idg])
            elif isinstance(definition_list, dict):
                for name in definition_list:
                    if isinstance(definition_list[name][0], list):
                        for ele in definition_list[name]:
                            self.define_group(group_type, name, ele)
                    else:
                        self.define_group(group_type, name,
                                          definition_list[name])

    def all_separate_groups(self, group_type):
        """Separate all possible types of twobodys, threebodys, manybody.
        One type per group.

        Args:
            group_type (str): "specie", "twobody", "threebody", "cut3b", "manybody"

        """
        nspec = len(self.all_group_names['specie'])
        if nspec < 1:
            raise RuntimeError("the specie group has to be defined in advance")
        if group_type in self.all_group_types:
            # TO DO: the two blocks below can be replace by some upper triangle operation

            # generate all possible combination of group
            ele_grid = self.all_group_names['specie']
            grid = np.meshgrid(*[ele_grid]*ParameterHelper.ndim[group_type])
            grid = np.array(grid).T.reshape(-1,
                                            ParameterHelper.ndim[group_type])

            # remove the redundant groups
            allgroup = []
            for group in grid:
                exist = False
                set_list_group = set(list(group))
                for prev_group in allgroup:
                    if set(prev_group) == set_list_group:
                        exist = True
                if not exist:
                    allgroup += [list(group)]

            # define the group
            tid = 0
            for group in allgroup:
                self.define_group(group_type, f'{group_type}{tid}',
                                  group)
                tid += 1
        else:
            self.logger.info(f"{group_type} will be ignored")

    def fill_in_parameters(self, group_type, random=False, ones=False, universal=False):
        """Separate all possible types of twobodys, threebodys, manybody.
        One type per group. And fill in either universal ls and sigma from
        pre-defined parameters from set_parameters("sigma", ..) and set_parameters("ls", ..)
        or random parameters if random is True.

        Args:
            group_type (str): "specie", "twobody", "threebody", "cut3b", "manybody"
            definition_list (list, dict): list of elements

        """
        nspec = len(self.all_group_names['specie'])
        if nspec < 1:
            raise RuntimeError("the specie group has to be defined in advance")
        if random:
            for group_name in self.all_group_names[group_type]:
                self.set_parameters(group_name, parameters=nprandom(2))
        elif ones:
            for group_name in self.all_group_names[group_type]:
                self.set_parameters(group_name, parameters=np.ones(2))
        elif universal:
            for group_name in self.all_group_names[group_type]:
                self.set_parameters(group_name,
                                    parameters=[self.universal['sigma'],
                                                self.universal['lengthscale']])

    def define_group(self, group_type, name, element_list, parameters=None, atomic_str=False):
        """Define specie/twobody/threebody/3b cutoff/manybody group

        Args:
            group_type (str): "specie", "twobody", "threebody", "cut3b", "manybody"
            name (str): the name use for indexing. can be anything but "*"
            element_list (list): list of elements
            parameters (list): corresponding parameters for this group
            atomic_str (bool): whether the elements in element_list are
                specified by group names or periodic table element names.

        The function is helped to define different groups for specie/twobody/threebody
        /3b cutoff/manybody terms. This function can be used for many times.
        The later one always overrides the former one.

        The name of the group has to be unique string (but not "*"), that
        define a group of species or twobodys, etc. If the same name is used,
        in two function calls, the definitions of the group will be merged.
        Both calls will be effective.

        element_list has to be a list of atomic elements, or a list of
        specie group names (which should be defined in previous calls), or "*".
        "*" will loop the function over all previously defined species.
        It has to be two elements for twobody/3b cutoff/manybody term, or
        three elements for threebody. For specie group definition, it can be
        as many elements as you want.

        If multiple define_group calls have conflict with element, the later one
        has higher priority. For example, twobody 1-2 are defined as group1 in
        the first call, and as group2 in the second call. In the end, the twobody
        will be left as group2.

        Example 1:

        >>> define_group('specie', 'water', ['H', 'O'])
        >>> define_group('specie', 'salt', ['Cl', 'Na'])

        They define H and O to be group water, and Na and Cl to be group salt.

        Example 2.1:

        >>> define_group('twobody', 'in-water', ['H', 'H'], atomic_str=True)
        >>> define_group('twobody', 'in-water', ['H', 'O'], atomic_str=True)
        >>> define_group('twobody', 'in-water', ['O', 'O'], atomic_str=True)

        Example 2.2:

        >>> define_group('twobody', 'in-water', ['water', 'water'])

        The 2.1 is equivalent to 2.2.

        Example 3.1:

        >>> define_group('specie', '1', ['H'])
        >>> define_group('specie', '2', ['O'])
        >>> define_group('twobody', 'Hgroup', ['H', 'H'], atomic_str=True)
        >>> define_group('twobody', 'Hgroup', ['H', 'O'], atomic_str=True)
        >>> define_group('twobody', 'OO', ['O', 'O'], atomic_str=True)

        Example 3.2:

        >>> define_group('specie', '1', ['H'])
        >>> define_group('specie', '2', ['O'])
        >>> define_group('twobody', 'Hgroup', ['H', '*'], atomic_str=True)
        >>> define_group('twobody', 'OO', ['O', 'O'], atomic_str=True)

        Example 3.3:

        >>> list_groups('specie', ['H', 'O'])
        >>> define_group('twobody', 'Hgroup', ['H', '*'])
        >>> define_group('twobody', 'OO', ['O', 'O'])

        Example 3.4:

        >>> list_groups('specie', ['H', 'O'])
        >>> define_group('twobody', 'OO', ['*', '*'])
        >>> define_group('twobody', 'Hgroup', ['H', '*'])

        3.1 to 3.4 are all equivalent.
        """

        if name == '*' and group_type == 'specie':
            name = 'allspecie'
            element_list = ['H']
        elif name == '*':
            raise ValueError("* is reserved for substitution, cannot be used "
                             "as a group name")

        if group_type != 'specie':

            assert len(element_list) == ParameterHelper.ndim[group_type]

            # Check all the other group_type to
            exclude_list = deepcopy(self.all_types)
            ide = exclude_list.index(group_type)
            exclude_list.pop(ide)

            for gt in exclude_list:
                if name in self.all_group_names[gt]:
                    raise ValueError("group name has to be unique across all types. "
                                     f"{name} is found in type {gt}")

        if name in self.all_group_names[group_type]:
            groupid = self.all_group_names[group_type].index(name)
        else:
            groupid = self.n[group_type]
            self.all_group_names[group_type].append(name)
            self.groups[group_type].append([])
            self.n[group_type] += 1

        if group_type == 'specie':
            for ele in element_list:
                assert ele not in self.all_members['specie'], \
                    "The element has already been defined"
                self.groups['specie'][groupid].append(ele)
                self.all_members['specie'].append(ele)
                self.logger.debug(
                    f"Element {ele} will be defined as group {name}")
        else:
            if len(self.all_group_names['specie']) == 0:
                raise RuntimeError("The atomic species have to be"
                                   "defined in advance")

            # first translate element/group name to group name
            group_name_list = []
            if atomic_str:
                for ele_name in element_list:
                    if ele_name == "*":
                        gid += ["*"]
                    else:
                        for idx in range(self.n['specie']):
                            group_name = self.all_group_names['specie'][idx]
                            if ele_name in self.groups['specie'][idx]:
                                group_name_list += [group_name]
                                self.logger.debug(f"Element {ele_name} is used for "
                                                  f"definition, but the whole group "
                                                  f"{group_name} is affected")
            else:
                group_name_list = element_list

            if "*" not in group_name_list:

                gid = []
                for ele_name in group_name_list:
                    gid += [self.all_group_names['specie'].index(ele_name)]

                for ele in self.all_members[group_type]:
                    if set(gid) == set(ele):
                        self.logger.debug(
                            f"the definition of {group_type} {ele} will be overriden")
                self.groups[group_type][groupid].append(gid)
                self.all_members[group_type].append(gid)
                self.logger.debug(
                    f"{group_type} {gid} will be defined as group {name}")
                if parameters is not None:
                    self.set_parameters(name, parameters)
            else:
                one_star_less = deepcopy(group_name_list)
                idstar = group_name_list.index('*')
                one_star_less.pop(idstar)
                for sub in self.all_group_names['specie']:
                    self.logger.debug(f"{sub}, {one_star_less}")
                    self.define_group(group_type, name,
                                      one_star_less + [sub], parameters=parameters,
                                      atomic_str=False)

    def find_group(self, group_type, element_list, atomic_str=False):
        """ find the group that contains the input pair

        Args:
            group_type (str): species, twobody, threebody, cut3b, manybody
            element_list (list): list of elements for a pair/triplet/coordination-pair
            atomic_str (bool): whether the elements in element_list are
                specified by group names or periodic table element names.

        Return:
            name (str):
        """

        # remember the later command override the earlier ones
        if group_type == 'specie':
            if not isinstance(element_list, str):
                self.logger.debug("for element, it has to be a string")
                return None
            name = None
            for igroup in range(self.n['specie']):
                gname = self.all_group_names[group_type][igroup]
                allspec = self.groups[group_type][igroup]
                if element_list in allspec:
                    name = gname
            if name is None:
                self.logger.debug("cannot find the group")
            return name
        else:
            if "*" in element_list:
                self.logger.debug("* cannot be used for find")
                return None
            gid = []
            for ele_name in element_list:
                gid += [self.all_group_names['specie'].index(ele_name)]
            name = None
            for igroup in range(self.n[group_type]):
                gname = self.all_group_names[group_type][igroup]
                for ele in self.groups[group_type][igroup]:
                    if set(gid) == set(ele):
                        name = gname
            self.logger.debug(f"find the group {name}")
            return name

    def set_parameters(self, name, parameters, opt=True):
        """Set the parameters for certain group

        Args:
            name (str): name of the patermeters
            parameters (list): the sigma, lengthscale, and cutoff of each group.
            opt (bool, list): whether to optimize the parameter or not

        The name of parameters can be the group name previously defined in
        define_group or list_groups function. Aside from the group name,
        ``noise``, ``cutoff_twobody``, ``cutoff_threebody``, and
        ``cutoff_manybody`` are reserved for noise parmater
        and universal cutoffs, while ``sigma`` and ``lengthscale`` are
        reserved for universal signal variances and length scales.

        The parameter should be a list of 2-3 elements, for sigma,
        lengthscale (and cutoff if the third one is defined).

        The optimization flag can be a single bool, which apply to all
        parameters, or list of bools that apply to each parameter.
        """

        if name == 'noise':
            self.noise = parameters
            self.opt['noise'] = opt
            self.logger.debug(f"noise will be set as "
                              f"{parameters} and opt {opt}")
            return
        elif name == 'energy_noise':
            self.energy_noise = parameters
            self.opt['energy_noise'] = opt
            self.logger.debug(f"energy_noise will be set as "
                              f"{parameters} and opt {opt}")
            return
        elif 'cutoff' in name:
            self.universal[name] = parameters
            self.logger.debug(f"universal cutoff {name} will be set as "
                              f"{parameters}")
            return
        elif name in ['sigma', 'lengthscale']:
            self.universal[name] = parameters
            self.opt[name] = opt
            self.logger.debug(f"universal {name} will be set as "
                              f"{parameters} and optimized {opt}")
            return

        if isinstance(opt, bool):
            opt = [opt]*2

        for cut_name in self.cutoff_types:
            if cut_name in name:
                self.all_cutoff[name] = float(parameters)
                self.logger.debug(f"Cutoff for group {name} will be set as "
                                  f"{parameters}")
                return

        if name in self.sigma:
            self.logger.debug(
                f"the sig, ls of group {name} is overriden")
        self.sigma[name] = parameters[0]
        self.ls[name] = parameters[1]
        self.opt[name+'sig'] = opt[0]
        self.opt[name+'ls'] = opt[1]
        self.logger.debug(f"ParameterHelper for group {name} will be set as "
                          f"sig={parameters[0]} ({opt[0]}) "
                          f"ls={parameters[1]} ({opt[1]})")
        if len(parameters) > 2:
            if name in self.all_cutoff:
                self.logger.debug(
                    f"the cutoff of group {name} is overriden")
            self.all_cutoff[name] = parameters[2]
            self.logger.debug(f"Cutoff for group {name} will be set as "
                              f"{parameters[2]}")

    def set_constraints(self, name, opt):
        """Set the parameters for certain group

        Args:
            name (str): name of the patermeters
            opt (bool, list): whether to optimize the parameter or not

        The name of parameters can be the group name previously defined in
        define_group or list_groups function. Aside from the group name,
        ``noise``, ``cutoff_twobody``, ``cutoff_threebody``, and
        ``cutoff_manybody`` are reserved for noise parmater
        and universal cutoffs, while ``sigma`` and ``lengthscale`` are
        reserved for universal signal variances and length scales.

        The optimization flag can be a single bool, which apply to all
        parameters under that name, or list of bools that apply to each
        parameter.
        """

        if name == 'noise':
            self.opt['noise'] = opt
            self.logger.debug(f"noise opt is set to"
                              f"{opt}")
            return

        if isinstance(opt, bool):
            opt = [opt, opt, opt]

        for cutname in self.cutoff_types:
            if cutname in name:
                return

        if name in self.sigma:
            self.logger.debug(
                f"the opt setting of group {name} is overriden")
        self.opt[name+'sig'] = opt[0]
        self.opt[name+'ls'] = opt[1]
        self.logger.debug(f"ParameterHelper for group {name} will be set as "
                          f"sig {opt[0]} "
                          f"ls {opt[1]}")

    def summarize_group(self, group_type):
        """Sort and combine all the previous definition to internal varialbes

        Args:
            group_type (str): species, twobody, threebody, cut3b, manybody
        """

        aeg = self.all_group_names[group_type]
        nspecie = self.n['specie']

        # specie need special sorting
        if group_type == "specie":

            self.nspecie = nspecie
            self.specie_mask = np.ones(118, dtype=np.int)*(nspecie-1)

            # mark the species_mask with atom type
            # default is nspecie-1
            for idt in range(self.nspecie):
                for ele in self.groups['specie'][idt]:
                    atom_n = element_to_Z(ele)
                    if atom_n >= len(self.specie_mask):
                        new_mask = np.ones(atom_n, dtype=np.int)*(nspecie-1)
                        new_mask[:len(self.specie_mask)] = self.specie_mask
                        self.species_mask = new_mask
                    self.specie_mask[atom_n] = idt
                    self.logger.debug(f"elemtn {ele} is defined as type {idt} with name "
                                      f"{aeg[idt]}")
            self.logger.debug(
                f"All the remaining elements are left as type {idt}")

        elif group_type in self.all_group_types:

            if self.n[group_type] == 0:
                self.logger.debug(f"{group_type} is not defined. Skipped")
                return

            if (group_type not in self.kernels) and \
                    (group_type in ParameterHelper.all_kernel_types):
                self.kernels.append(group_type)

            self.mask[group_type] = np.ones(
                nspecie**ParameterHelper.ndim[group_type], dtype=np.int) *\
                (self.n[group_type]-1)

            self.hyps_sig[group_type] = []
            self.hyps_ls[group_type] = []
            all_opt_sig = []
            all_opt_ls = []

            for idt in range(self.n[group_type]):
                name = aeg[idt]
                for ele_list in self.groups[group_type][idt]:
                    # generate all possible permutation
                    perms = list(permutations(ele_list))
                    for ele_list in perms:
                        mask_id = 0
                        for ele in ele_list:
                            mask_id += ele
                            mask_id *= nspecie
                        mask_id = mask_id // nspecie
                        self.mask[group_type][mask_id] = idt
                    def_str = "-".join(map(str, self.groups['specie']))
                    self.logger.debug(f"{group_type} {def_str} is defined as type {idt} "
                                      f"with name {name}")

                if group_type not in self.cutoff_types:
                    sig = self.sigma.get(name, -1)
                    opt_sig = self.opt.get(name+'sig', True)
                    if sig == -1:
                        sig = self.sigma.get(group_type, -1)
                        opt_sig = self.opt.get(group_type+'sig', True)
                    if sig == -1:
                        sig = self.universal.get('sigma', -1)
                        opt_sig = self.opt.get('sigma', True)

                    ls = self.ls.get(name, -1)
                    opt_ls = self.opt.get(name+'ls', True)
                    if ls == -1:
                        ls = self.ls.get(group_type, -1)
                        opt_ls = self.opt.get(group_type+'ls', True)
                    if ls == -1:
                        ls = self.universal.get('lengthscale', -1)
                        opt_ls = self.opt.get('lengthscale', True)

                    if sig < 0 or ls < 0:
                        self.logger.error(f"hyper parameters for group {name}"
                                          " is not defined")
                        raise RuntimeError
                    self.hyps_sig[group_type] += [sig]
                    self.hyps_ls[group_type] += [ls]
                    all_opt_sig += [opt_sig]
                    all_opt_ls += [opt_ls]
                    self.logger.debug(f"   using hyper-parameters of {sig:6.2g} "
                                      f"{ls:6.2g} {opt_sig} {opt_ls}")
            self.hyps_opt[group_type] = all_opt_sig + all_opt_ls
            self.logger.debug(
                f"All the remaining elements are left as type {idt}")

            # sort out the cutoffs
            if group_type in self.cutoff_types:
                universal_cutoff = self.universal.get(
                    'cutoff_'+self.cutoff_types[group_type], 0)
            else:
                universal_cutoff = self.universal.get('cutoff_'+group_type, 0)

            allcut = []
            alldefine = True
            for idt in range(self.n[group_type]):
                if aeg[idt] in self.all_cutoff:
                    allcut += [self.all_cutoff[aeg[idt]]]
                else:
                    alldefine = False
                    self.logger.info(f"{aeg[idt]} cutoff is not defined. "
                                     "it's going to use the universal cutoff.")
            if group_type not in self.cutoff_types_values:

                if len(allcut) > 0:
                    if universal_cutoff <= 0:
                        universal_cutoff = np.max(allcut)
                        self.logger.info(f"universal cutoff for "
                                         f"{group_type} is defined as zero! reset it to {universal_cutoff}")

                    self.cutoff_list[group_type] = []
                    for idt in range(self.n[group_type]):
                        self.cutoff_list[group_type] += [
                            self.all_cutoff.get(aeg[idt], universal_cutoff)]
                    self.cutoff_list[group_type] = np.array(
                        self.cutoff_list[group_type], dtype=float)

                    max_cutoff = np.max(self.cutoff_list[group_type])

                    # update the universal cutoff to make it higher than
                    if alldefine:
                        universal_cutoff = max_cutoff
                        self.logger.info(f"universal cutoff is updated to "
                                         f"{universal_cutoff}")
                    elif not np.any(self.cutoff_list[group_type]-max_cutoff):
                        # if not all the cutoffs are defined separately
                        # and they are all the same value
                        del self.cutoff_list[group_type]
                        universal_cutoff = max_cutoff
                        if group_type in self.cutoff_types:
                            self.n[group_type] = 0
                        self.logger.info(f"universal cutoff is updated to "
                                         f"{universal_cutoff}")

            else:
                if universal_cutoff <= 0 and len(allcut) > 0:
                    universal_cutoff = np.max(allcut)
                    self.logger.info(f"threebody universal cutoff is updated to"
                                     f"{universal_cutoff}, but the separate definitions will"
                                     "be ignored")

            if universal_cutoff > 0:
                if group_type in self.cutoff_types:
                    keyname = 'cutoff_'+self.cutoff_types[group_type]
                else:
                    keyname = 'cutoff_'+group_type
                self.universal[keyname] = universal_cutoff
            else:
                self.logger.error(f"cutoffs for {group_type} is undefined")
                raise RuntimeError

        else:
            pass

    def as_dict(self):
        """Dictionary representation of the mask. The output can be used for AtomicEnvironment
        or the GaussianProcess
        """

        # sort out all the definitions and resolve conflicts
        # cut3b has to be summarize before threebody
        # because the universal threebody cutoff is checked
        # at the end of threebody search

        self.summarize_group('specie')
        for ktype in self.cutoff_types:
            self.summarize_group(ktype)
        for ktype in ParameterHelper.additional_groups:
            self.summarize_group(ktype)
        for ktype in ParameterHelper.all_kernel_types:
            self.summarize_group(ktype)

        hyps_mask = {}
        cutoff_dict = {}

        hyps_mask['nspecie'] = self.n['specie']
        if self.n['specie'] > 1:
            hyps_mask['specie_mask'] = self.specie_mask

        hyps = []
        hyp_labels = []
        opt = []
        for group in self.kernels:

            hyps_mask['n'+group] = self.n[group]
            hyps_mask[group+'_start'] = len(hyps)
            hyps += [self.hyps_sig[group]]
            hyps += [self.hyps_ls[group]]
            hyps = list(np.hstack(hyps))
            opt += [self.hyps_opt[group]]
            cutoff_dict[group] = self.universal['cutoff_'+group]

            if self.n[group] > 1:
                hyps_mask[group+'_mask'] = self.mask[group]
                # check parameters
                aeg = self.all_group_names[group]
                for idt in range(self.n[group]):
                    hyp_labels += ['Signal Std '+aeg[idt]]
                for idt in range(self.n[group]):
                    hyp_labels += ['Length Scale '+aeg[idt]]
            else:
                hyp_labels += ['Signal Std '+group]
                hyp_labels += ['Length Scale '+group]

            if group in self.cutoff_list:
                hyps_mask[group+'_cutoff_list'] = self.cutoff_list[group]

        for cutoff_name in self.cutoff_types:
            if self.n.get(cutoff_name, 0) >= 1:
                hyps_mask['n'+cutoff_name] = self.n[cutoff_name]
                hyps_mask[cutoff_name+'_mask'] = self.mask[cutoff_name]
                hyps_mask[self.cutoff_types[cutoff_name] +
                          '_cutoff_list'] = self.cutoff_list[cutoff_name]

        hyps_mask['train_noise'] = self.opt['noise']
        hyps_mask['energy_noise'] = self.energy_noise

        opt += [self.opt['noise']]
        hyp_labels += ['Noise std']
        hyps += [self.noise]
        hyps = np.hstack(hyps)
        opt = np.hstack(opt)

        # handle partial optimization if any constraints are defined
        if not opt.all():
            nhyps = len(hyps)
            hyps_mask['original_hyps'] = hyps
            hyps_mask['original_labels'] = hyp_labels
            mapping = []
            new_labels = []
            for i in range(nhyps):
                if opt[i]:
                    mapping += [i]
                    new_labels += [hyp_labels[i]]
            newhyps = hyps[mapping]
            hyps_mask['map'] = np.array(mapping, dtype=np.int)
        elif opt.any():
            newhyps = hyps
            new_labels = hyp_labels
        else:
            raise RuntimeError("hyps has length zero."
                               "at least one component of the hyper-parameters"
                               "should be allowed to be optimized. \n")

        if self.n['specie'] < 2:
            self.logger.debug(
                "only one type of elements was defined. Please use multihyps=False")

        hyps_mask['kernels'] = self.kernels
        hyps_mask['kernel_name'] = "+".join(hyps_mask['kernels'])
        hyps_mask['cutoffs'] = cutoff_dict
        hyps_mask['hyps'] = newhyps
        hyps_mask['hyp_labels'] = new_labels

        logging.debug(str(hyps_mask))

        return hyps_mask

    @staticmethod
    def from_dict(hyps_mask, verbose=False, init_spec=[]):
        """ convert dictionary mask to HM instance
        This function is not tested yet
        """

        Parameters.check_instantiation(hyps_mask['hyps'],
                                       hyps_mask['cutoffs'],
                                       hyps_mask['kernels'],
                                       hyps_mask)

        pm = ParameterHelper(verbose=verbose)

        nspecie = hyps_mask['nspecie']
        if nspecie > 1:
            max_species = np.max(hyps_mask['specie_mask'])
            specie_mask = hyps_mask['specie_mask']
            for i in range(max_species+1):
                elelist = np.where(specie_mask == i)[0]
                if len(elelist) > 0:
                    for ele in elelist:
                        if ele != 0:
                            elename = Z_to_element(ele)
                            if len(init_spec) > 0:
                                if elename in init_spec:
                                    pm.define_group(
                                        "specie", i, [elename])
                            else:
                                pm.define_group("specie", i, [elename])
        else:
            pm.define_group("specie", i, ['*'])

        for kernel in hyps_mask['kernels']+ParameterHelper.cutoff_types_keys:
            n = hyps_mask.get('n'+kernel, 0)
            if n >= 0:
                if kernel not in ParameterHelper.cutoff_types:
                    chyps, copt = Parameters.get_component_hyps(hyps_mask, kernel,
                                                                constraint=True, noise=False)
                    sig = chyps[0]
                    ls = chyps[1]
                    csig = copt[0]
                    cls = copt[1]
                    cutoff = hyps_mask['cutoffs'][kernel]
                    pm.set_parameters('cutoff_'+kernel, cutoff)
                    cutoff_list = hyps_mask.get(
                        f'{kernel}_cutoff_list', np.ones(len(sig))*cutoff)
                elif kernel in ParameterHelper.cutoff_types and n > 1:
                    cutoff_list = hyps_mask[ParameterHelper.cutoff_types[kernel]+'_cutoff_list']

                if n > 1:
                    all_specie = np.arange(nspecie)
                    all_comb = combinations_with_replacement(
                        all_specie, ParameterHelper.ndim[kernel])
                    for comb in all_comb:
                        mask_id = 0
                        for ele in comb:
                            mask_id += ele
                            mask_id *= nspecie
                        mask_id = mask_id // nspecie
                        ttype = hyps_mask[f'{kernel}_mask'][mask_id]
                        pm.define_group(f"{kernel}", f"{kernel}{ttype}", comb)

                        if (kernel not in ParameterHelper.cutoff_types) and \
                                (kernel not in ParameterHelper.cutoff_types_values):
                            pm.set_parameters(f"{kernel}{ttype}", [sig[ttype], ls[ttype], cutoff_list[ttype]],
                                              opt=[csig[ttype], cls[ttype]])
                        elif kernel in ParameterHelper.cutoff_types_values:
                            pm.set_parameters(f"{kernel}{ttype}", [sig[ttype], ls[ttype]],
                                              opt=[csig[ttype], cls[ttype]])
                        else:
                            pm.set_parameters(
                                f"{kernel}{ttype}", cutoff_list[ttype])
                else:
                    pm.define_group(kernel, kernel, [
                                    '*']*ParameterHelper.ndim[kernel])
                    if kernel not in ParameterHelper.cutoff_types_keys:
                        pm.set_parameters(kernel, parameters=np.hstack(
                            [sig, ls, cutoff]), opt=copt)
                    else:
                        pm.set_parameters(kernel, parameters=cutoff)

        hyps = Parameters.get_hyps(hyps_mask)
        pm.set_parameters('noise', hyps[-1])

        if 'cutoffs' in hyps_mask:
            cutoffs = hyps_mask['cutoffs']
            for k in cutoffs:
                pm.set_parameters(f"cutoff_{k}", cutoffs[k])

        return pm
